##
# Copyright 2009-2023 Ghent University
#
# This file is part of EasyBuild,
# originally created by the HPC team of Ghent University (http://ugent.be/hpc/en),
# with support of Ghent University (http://ugent.be/hpc),
# the Flemish Supercomputer Centre (VSC) (https://www.vscentrum.be),
# Flemish Research Foundation (FWO) (http://www.fwo.be/en)
# and the Department of Economy, Science and Innovation (EWI) (http://www.ewi-vlaanderen.be/en).
#
# https://github.com/easybuilders/easybuild
#
# EasyBuild is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation v2.
#
# EasyBuild is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with EasyBuild.  If not, see <http://www.gnu.org/licenses/>.
##
"""
EasyBuild support for Boost, implemented as an easyblock

@author: Stijn De Weirdt (Ghent University)
@author: Dries Verdegem (Ghent University)
@author: Kenneth Hoste (Ghent University)
@author: Pieter De Baets (Ghent University)
@author: Jens Timmerman (Ghent University)
@author: Ward Poelmans (Ghent University)
@author: Petar Forai (IMP/IMBA)
@author: Luca Marsella (CSCS)
@author: Guilherme Peretti-Pezzi (CSCS)
@author: Joachim Hein (Lund University)
@author: Michele Dolfi (ETH Zurich)
@author: Simon Branford (University of Birmingham)
"""
from distutils.version import LooseVersion
import fileinput
import glob
import os
import re
import sys

import easybuild.tools.toolchain as toolchain
from easybuild.framework.easyblock import EasyBlock
from easybuild.framework.easyconfig import CUSTOM
from easybuild.tools.build_log import EasyBuildError
from easybuild.tools.config import ERROR
from easybuild.tools.filetools import apply_regex_substitutions, read_file, symlink, which, write_file
from easybuild.tools.modules import get_software_root, get_software_version
from easybuild.tools.run import run_cmd
from easybuild.tools.systemtools import AARCH64, POWER, UNKNOWN
from easybuild.tools.systemtools import get_cpu_architecture, get_glibc_version, get_shared_lib_ext


class EB_Boost(EasyBlock):
    """Support for building Boost."""

    def __init__(self, *args, **kwargs):
        """Initialize Boost-specific variables."""
        super(EB_Boost, self).__init__(*args, **kwargs)

        self.pyvers = []

        if LooseVersion(self.version) >= LooseVersion("1.71.0"):
            self.bjamcmd = 'b2'
        else:
            self.bjamcmd = 'bjam'

    @staticmethod
    def extra_options():
        """Add extra easyconfig parameters for Boost."""
        extra_vars = {
            'boost_mpi': [False, "Build mpi boost module", CUSTOM],
            'boost_multi_thread': [None, "Build boost with multi-thread option (DEPRECATED)", CUSTOM],
            'tagged_layout': [None, "Build with tagged layout on library names, default from version 1.69.0", CUSTOM],
            'single_threaded': [None, "Also build single threaded libraries, requires tagged_layout, "
                                      "default from version 1.69.0", CUSTOM],
            'toolset': [None, "Toolset to use for Boost configuration ('--with-toolset' for bootstrap.sh)", CUSTOM],
            'build_toolset': [None, "Toolset to use for Boost compilation "
                                    "('toolset' for b2, default calculated from toolset)", CUSTOM],
            'mpi_launcher': [None, "Launcher to use when running MPI regression tests", CUSTOM],
            'only_python_bindings': [False, "Only install Boost.Python library providing Python bindings", CUSTOM],
            'use_glibcxx11_abi': [None, "Use the GLIBCXX11 ABI", CUSTOM],
        }
        return EasyBlock.extra_options(extra_vars)

    def patch_step(self):
        """Patch Boost source code before building."""
        super(EB_Boost, self).patch_step()

        # TIME_UTC is also defined in recent glibc versions, so we need to rename it for old Boost versions (<= 1.49)
        glibc_version = get_glibc_version()
        old_glibc = glibc_version is not UNKNOWN and LooseVersion(glibc_version) > LooseVersion("2.15")
        if old_glibc and LooseVersion(self.version) <= LooseVersion("1.49.0"):
            self.log.info("Patching because the glibc version is too new")
            files_to_patch = ["boost/thread/xtime.hpp"] + glob.glob("libs/interprocess/test/*.hpp")
            files_to_patch += glob.glob("libs/spirit/classic/test/*.cpp") + glob.glob("libs/spirit/classic/test/*.inl")
            for patchfile in files_to_patch:
                try:
                    for line in fileinput.input("%s" % patchfile, inplace=1, backup='.orig'):
                        line = re.sub(r"TIME_UTC", r"TIME_UTC_", line)
                        sys.stdout.write(line)
                except IOError as err:
                    raise EasyBuildError("Failed to patch %s: %s", patchfile, err)

    def prepare_step(self, *args, **kwargs):
        """Prepare build environment."""

        super(EB_Boost, self).prepare_step(*args, **kwargs)

        # keep track of Python version(s) used during installation,
        # so we can perform a complete sanity check
        if get_software_root('Python'):
            self.pyvers.append(get_software_version('Python'))

    def configure_step(self):
        """Configure Boost build using custom tools"""

        # boost_multi_thread is deprecated
        if self.cfg['boost_multi_thread'] is not None:
            self.log.deprecated("boost_multi_thread has been replaced by tagged_layout. "
                                "We build with tagged layout and both single and multi threading libraries "
                                "from version 1.69.0.", '5.0')
            self.cfg['tagged_layout'] = True

        # mpi sanity check
        if self.cfg['boost_mpi'] and not self.toolchain.options.get('usempi', None):
            raise EasyBuildError("When enabling building boost_mpi, also enable the 'usempi' toolchain option.")

        # generate config depending on compiler used
        toolset = self.cfg['toolset']
        if toolset is None:
            if self.toolchain.comp_family() == toolchain.INTELCOMP:
                toolset = 'intel-linux'
            elif self.toolchain.comp_family() == toolchain.GCC:
                toolset = 'gcc'
            else:
                raise EasyBuildError("Unknown compiler used, don't know what to specify to --with-toolset, aborting.")

        cmd = "%s ./bootstrap.sh --with-toolset=%s --prefix=%s %s"
        tup = (self.cfg['preconfigopts'], toolset, self.installdir, self.cfg['configopts'])
        run_cmd(cmd % tup, log_all=True, simple=True)

        # Use build_toolset if specified or the bootstrap toolset without the OS suffix
        self.toolset = self.cfg['build_toolset'] or re.sub('-linux$', '', toolset)

        user_config = []

        # Explicitely set the compiler path to avoid B2 checking some standard paths like /opt
        cxx = os.getenv('CXX')
        if cxx:
            cxx = which(cxx, on_error=ERROR)
            # Remove default toolset config which may lead to duplicate toolsets (e.g. for intel-linux)
            apply_regex_substitutions('project-config.jam', [('using %s ;' % toolset, '')])
            # Add our toolset config with no version and full path to compiler
            user_config.append("using %s : : %s ;" % (self.toolset, cxx))

        if self.cfg['boost_mpi']:

            # configure the boost mpi module
            # http://www.boost.org/doc/libs/1_47_0/doc/html/mpi/getting_started.html
            # let Boost.Build know to look here for the config file

            # Check if using a Cray toolchain and configure MPI accordingly
            if self.toolchain.toolchain_family() == toolchain.CRAYPE:
                if self.toolchain.PRGENV_MODULE_NAME_SUFFIX == 'gnu':
                    craympichdir = os.getenv('CRAY_MPICH2_DIR')
                    craygccversion = os.getenv('GCC_VERSION')
                    # We configure the gcc toolchain below, so make sure the EC doesn't use another toolset
                    if self.toolset != 'gcc':
                        raise EasyBuildError("For the cray toolchain the 'gcc' toolset must be used.")
                    # Remove the previous "using gcc" line add above (via self.toolset) if present
                    user_config = [x for x in user_config if not x.startswith('using gcc :')]
                    user_config.extend([
                        'local CRAY_MPICH2_DIR =  %s ;' % craympichdir,
                        'using gcc ',
                        ': %s' % craygccversion,
                        ': CC ',
                        ': <compileflags>-I$(CRAY_MPICH2_DIR)/include ',
                        r'  <linkflags>-L$(CRAY_MPICH2_DIR)/lib \ ',
                        '; ',
                        'using mpi ',
                        ': CC ',
                        ': <find-shared-library>mpich ',
                        ': %s' % self.cfg['mpi_launcher'],
                        ';',
                        '',
                    ])
                else:
                    raise EasyBuildError("Bailing out: only PrgEnv-gnu supported for now")
            else:
                user_config.append("using mpi : %s ;" % os.getenv("MPICXX"))

        write_file('user-config.jam', '\n'.join(user_config), append=True)

    def build_step(self):
        """Build Boost with bjam tool."""

        self.bjamoptions = " --prefix=%s --user-config=user-config.jam" % self.installdir
        if 'toolset=' not in self.cfg['buildopts']:
            self.bjamoptions += " toolset=" + self.toolset

        cxxflags = os.getenv('CXXFLAGS')
        # only disable -D_GLIBCXX_USE_CXX11_ABI if use_glibcxx11_abi was explicitly set to False
        # None value is the default, which corresponds to default setting (=1 since GCC 5.x)
        if self.cfg['use_glibcxx11_abi'] is not None:
            cxxflags += ' -D_GLIBCXX_USE_CXX11_ABI='
            if self.cfg['use_glibcxx11_abi']:
                cxxflags += '1'
            else:
                cxxflags += '0'
        if cxxflags:
            self.bjamoptions += " cxxflags='%s'" % cxxflags
        ldflags = os.getenv('LDFLAGS')
        if ldflags:
            self.bjamoptions += " linkflags='%s'" % ldflags

        # specify path for bzip2/zlib if module is loaded
        for lib in ["bzip2", "zlib"]:
            libroot = get_software_root(lib)
            if libroot:
                self.bjamoptions += " -s%s_INCLUDE=%s/include" % (lib.upper(), libroot)
                self.bjamoptions += " -s%s_LIBPATH=%s/lib" % (lib.upper(), libroot)

        if self.cfg['parallel']:
            self.paracmd = "-j %s" % self.cfg['parallel']
        else:
            self.paracmd = ''

        # Add list of default library settings from project-config (created by configure step)
        # Required because any --with-* or --without-* overwrites this entirely
        project_config = read_file('project-config.jam')
        libraries = re.search(r'libraries = (.*) ;', project_config)
        if libraries:
            self.bjamoptions += libraries.group(1)

        if self.cfg['only_python_bindings']:
            # magic incantation to only install Boost Python bindings is... --with-python
            # see http://boostorg.github.io/python/doc/html/building/installing_boost_python_on_your_.html
            self.bjamoptions += " --with-python"

        if LooseVersion(self.version) >= LooseVersion("1.69.0"):
            # As of 1.69.0 we build with layout tagged and both single and multi threading
            # Linking default libraries to multi-threaded versions.
            if self.cfg['tagged_layout'] is None:
                self.cfg['tagged_layout'] = True
            if self.cfg['single_threaded'] is None:
                self.cfg['single_threaded'] = True

        # Default threading since at least 1.47.0 is multi with system layout

        if self.cfg['tagged_layout']:
            layout = "tagged"
        else:
            layout = "system"

        if self.cfg['single_threaded']:
            if not self.cfg['tagged_layout']:
                raise EasyBuildError("Singled threaded build requires tagged layout.")
            threading = "single,multi"
        else:
            threading = "multi"

        self.bjamoptions += " threading=" + threading + " --layout=" + layout

        if not self.cfg['boost_mpi'] and not self.cfg['only_python_bindings']:
            # Default but avoids a warning. Building Boost.MPI is actually enabled by `using mpi` in the user-config
            # Note: Can't use both --with-* and --without-*
            self.bjamoptions += " --without-mpi"

        self.log.info("Building Boost libraries")
        # build with specified options
        cmd = ' '.join([
            self.cfg['prebuildopts'],
            os.path.join('.', self.bjamcmd),
            self.bjamoptions,
            self.paracmd,
            self.cfg['buildopts'],
        ])
        run_cmd(cmd, log_all=True, simple=True)

    def install_step(self):
        """Install Boost by copying files to install dir."""

        # install boost libraries
        self.log.info("Installing Boost libraries")

        cmd = ' '.join([
            self.cfg['preinstallopts'],
            os.path.join('.', self.bjamcmd),
            self.bjamoptions,
            'install',
            self.paracmd,
            self.cfg['installopts'],
        ])
        run_cmd(cmd, log_all=True, simple=True)

        if self.cfg['tagged_layout']:
            if LooseVersion(self.version) >= LooseVersion("1.69.0") or not self.cfg['single_threaded']:
                # Link tagged multi threaded libs as the default libs
                lib_glob = 'lib*-mt*.*'
                mt_replace = re.compile(r'-[^.]*\.')
                for source_lib in glob.glob(os.path.join(self.installdir, 'lib', lib_glob)):
                    target_lib = mt_replace.sub('.', os.path.basename(source_lib))
                    symlink(os.path.basename(source_lib), os.path.join(self.installdir, 'lib', target_lib),
                            use_abspath_source=False)

    def sanity_check_step(self):
        """Custom sanity check for Boost."""
        shlib_ext = get_shared_lib_ext()

        custom_paths = {
            'files': [],
            'dirs': ['include/boost']
        }
        if self.cfg['tagged_layout']:
            lib_mt_suffix = '-mt'
            # Architecture tags introduced in 1.69.0
            if LooseVersion(self.version) >= LooseVersion("1.69.0"):
                if get_cpu_architecture() == AARCH64:
                    lib_mt_suffix += '-a64'
                elif get_cpu_architecture() == POWER:
                    lib_mt_suffix += '-p64'
                else:
                    lib_mt_suffix += '-x64'

        if self.cfg['only_python_bindings']:
            for pyver in self.pyvers:
                pymajorver, pyminorver = pyver.split('.')[:2]
                if LooseVersion(self.version) >= LooseVersion("1.67.0"):
                    suffix = '%s%s' % (pymajorver, pyminorver)
                elif int(pymajorver) >= 3:
                    suffix = pymajorver
                else:
                    suffix = ''
                custom_paths['files'].append(os.path.join('lib', 'libboost_python%s.%s' % (suffix, shlib_ext)))
                if self.cfg['tagged_layout']:
                    custom_paths['files'].append(
                        os.path.join('lib', 'libboost_python%s%s.%s' % (suffix, lib_mt_suffix, shlib_ext)))

        else:
            custom_paths['files'].append(os.path.join('lib', 'libboost_system.%s' % shlib_ext))

            if self.cfg['tagged_layout']:
                custom_paths['files'].append(os.path.join('lib', 'libboost_system%s.%s' % (lib_mt_suffix, shlib_ext)))
                custom_paths['files'].append(os.path.join('lib', 'libboost_thread%s.%s' % (lib_mt_suffix, shlib_ext)))

            if self.cfg['boost_mpi']:
                custom_paths['files'].append(os.path.join('lib', 'libboost_mpi.%s' % shlib_ext))
                if self.cfg['tagged_layout']:
                    custom_paths['files'].append(os.path.join('lib', 'libboost_mpi%s.%s' % (lib_mt_suffix, shlib_ext)))

        super(EB_Boost, self).sanity_check_step(custom_paths=custom_paths)

    def make_module_extra(self):
        """Set up a BOOST_ROOT environment variable to e.g. ease Boost handling by cmake"""
        txt = super(EB_Boost, self).make_module_extra()
        if not self.cfg['only_python_bindings']:
            txt += self.module_generator.set_environment('BOOST_ROOT', self.installdir)
        return txt



##
# Copyright 2009-2023 Ghent University
#
# This file is part of EasyBuild,
# originally created by the HPC team of Ghent University (http://ugent.be/hpc/en),
# with support of Ghent University (http://ugent.be/hpc),
# the Flemish Supercomputer Centre (VSC) (https://www.vscentrum.be),
# Flemish Research Foundation (FWO) (http://www.fwo.be/en)
# and the Department of Economy, Science and Innovation (EWI) (http://www.ewi-vlaanderen.be/en).
#
# https://github.com/easybuilders/easybuild
#
# EasyBuild is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation v2.
#
# EasyBuild is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with EasyBuild.  If not, see <http://www.gnu.org/licenses/>.
##
"""
Geant4 support, implemented as an easyblock.

@author: Stijn De Weirdt (Ghent University)
@author: Dries Verdegem (Ghent University)
@author: Kenneth Hoste (Ghent University)
@author: Pieter De Baets (Ghent University)
@author: Jens Timmerman (Ghent University)
"""

import os
import shutil
import re
from distutils.version import LooseVersion

import easybuild.tools.environment as env
from easybuild.framework.easyconfig import CUSTOM
from easybuild.easyblocks.generic.cmakemake import CMakeMake
from easybuild.tools.build_log import EasyBuildError
from easybuild.tools.modules import get_software_root
from easybuild.tools.run import run_cmd, run_cmd_qa
from easybuild.tools.filetools import copy_file, read_file, write_file


class EB_Geant4(CMakeMake):
    """
    Support for building Geant4.
    Note: Geant4 moved to a CMake-based build system as of version 9.4.
    """

    @staticmethod
    def extra_options():
        """
        Define extra options needed by Geant4
        """
        extra_vars = CMakeMake.extra_options()
        extra_vars.update({
            'G4ABLAVersion': [None, "G4ABLA version", CUSTOM],
            'G4NDLVersion': [None, "G4NDL version", CUSTOM],
            'G4EMLOWVersion': [None, "G4EMLOW version", CUSTOM],
            'PhotonEvaporationVersion': [None, "PhotonEvaporation version", CUSTOM],
            'G4RadioactiveDecayVersion': [None, "G4RadioactiveDecay version", CUSTOM],
        })
        # Requires out-of-source build
        extra_vars['separate_build_dir'][0] = True
        return extra_vars

    def __init__(self, *args, **kwargs):
        """Initialisation of custom class variables for Geant4."""
        super(EB_Geant4, self).__init__(*args, **kwargs)
        self.g4system = 'UNKNOWN'
        self.datadst = 'UNKNOWN'

    def configure_step(self):
        """
        Configure Geant4 build, either via CMake for versions more recent than 9.4,
        or using an interactive configuration procedure otherwise.
        """

        # Geant4 switched to a CMake build system in version 9.4
        if LooseVersion(self.version) >= LooseVersion("9.4"):
            super(EB_Geant4, self).configure_step()

        else:
            pwd = self.cfg['start_dir']
            dst = self.installdir
            clhepdir = get_software_root('CLHEP')
            cmd = "%s/Configure -E -build" % pwd

            self.qanda = {
                # questions and answers for version 9.1.p03
                "There exists a config.sh file. Shall I use it to set the defaults? [y]": "n",
                "Would you like to see the instructions? [n]": "",
                "[Type carriage return to continue]": "",
                "Definition of G4SYSTEM variable is Linux-g++. That stands for: 1) OS : Linux" \
                "2) Compiler : g++ To modify default settings, select number above (e.g. 2) " \
                "[Press [Enter] for default settings]": "2",
                "Which C++ compiler? [g++]": "$(GPP)",
                "Confirm your selection or set recommended 'g++'! [*]": "",
                "Definition of G4SYSTEM variable is Linux-icc. That stands for: 1) OS : Linux 2)" \
                "Compiler : icc To modify default settings, select number above (e.g. 2) " \
                "[Press [Enter] for default settings]": "",
                "Do you expect to run these scripts and binaries on multiple machines? [n]": "y",
                "Where is Geant4 source installed? [%s]" % pwd: "",
                "Specify the path where Geant4 libraries and source files should be installed." \
                " [%s]" % pwd: dst,
                "Do you want to copy all Geant4 headers in one directory? [n]": "y",
                "Please, specify default directory where ALL the Geant4 data is installed:" \
                "G4LEVELGAMMADATA: %(pwd)s/data/PhotonEvaporation2.0 G4RADIOACTIVEDATA: " \
                "%(pwd)s/data/RadioactiveDecay3.2 G4LEDATA: %(pwd)s/data/G4EMLOW5.1 G4NEUTRONHPDATA:    " \
                "%(pwd)s/data/G4NDL3.12 G4ABLADATA: %(pwd)s/data/G4ABLA3.0 You will be asked about " \
                "customizing these next. [%(pwd)s/data]" % {'pwd': pwd}: "%s/data" % dst,
                "Directory %s/data doesn't exist. Use that name anyway? [n]" % dst: "y",
                "Please, specify default directory where the Geant4 data is installed: " \
                "1) G4LEVELGAMMADATA: %(dst)s/data/PhotonEvaporation2.0 2) G4RADIOACTIVEDATA: " \
                "%(dst)s/data/RadioactiveDecay3.2 3) G4LEDATA: %(dst)s/data/G4EMLOW5.1 4) G4NEUTRONHPDATA: " \
                "%(dst)s/data/G4NDL3.12 5) G4ABLADATA: %(dst)s/data/G4ABLA3.0 To modify default settings, " \
                "select number above (e.g. 2) [Press [Enter] for default settings]" % {'dst': dst}: "",
                "Please, specify where CLHEP is installed: CLHEP_BASE_DIR: ": clhepdir,
                "Please, specify where CLHEP is installed: CLHEP_BASE_DIR: [%s]" % clhepdir: "",
                "You can customize paths and library name of you CLHEP installation: 1) CLHEP_INCLUDE_DIR: " \
                "%(clhepdir)s/include 2) CLHEP_LIB_DIR: %(clhepdir)s/lib 3) CLHEP_LIB: CLHEP To modify " \
                "default settings, select number above (e.g. 2) [Press [Enter] for default settings]" %
                {'clhepdir': clhepdir}: "",
                "By default 'static' (.a) libraries are built. Do you want to build 'shared' (.so) " \
                "libraries? [n]": "y",
                "You selected to build 'shared' (.so) libraries. Do you want to build 'static' (.a) " \
                "libraries too? [n]": "y",
                "Do you want to build 'global' compound libraries? [n]": "",
                "Do you want to compile libraries in DEBUG mode (-g)? [n]": "",
                "G4UI_NONE If this variable is set, no UI sessions nor any UI libraries are built. " \
                "This can be useful when running a pure batch job or in a user framework having its own " \
                "UI system. Do you want to set this variable ? [n]": "",
                "G4UI_BUILD_XAW_SESSION G4UI_USE_XAW Specifies to include and use the XAW interfaces in " \
                "the application to be built. The XAW (X11 Athena Widget set) extensions are required to " \
                "activate and build this driver. [n]": "",
                "G4UI_BUILD_XM_SESSION G4UI_USE_XM Specifies to include and use the XM Motif based user " \
                "interfaces. The XM Motif extensions are required to activate and build this driver. [n]": "",
                "G4VIS_NONE If this variable is set, no visualization drivers will be built or used. Do " \
                "you want to set this variable ? [n]": "n",
                "G4VIS_BUILD_OPENGLX_DRIVER G4VIS_USE_OPENGLX It is an interface to the de facto standard " \
                "3D graphics library, OpenGL. It is well suited for real-time fast visualization and " \
                "prototyping. The X11 version of the OpenGL libraries is required. [n]": "",
                "G4VIS_BUILD_OPENGLXM_DRIVER G4VIS_USE_OPENGLXM It is an interface to the de facto " \
                "standard 3D graphics library, OpenGL. It is well suited for real-time fast visualization " \
                "and prototyping. The X11 version of the OpenGL libraries and the Motif Xm extension is " \
                "required. [n]": "",
                "G4VIS_BUILD_DAWN_DRIVER G4VIS_USE_DAWN DAWN drivers are interfaces to the Fukui Renderer " \
                "DAWN. DAWN is a vectorized 3D PostScript processor suited to prepare technical high " \
                "quality outputs for presentation and/or documentation. [n]": "",
                "G4VIS_BUILD_OIX_DRIVER G4VIS_USE_OIX The OpenInventor driver is based on OpenInventor tech" \
                "nology for scientific visualization. The X11 version of OpenInventor is required. [n]": "",
                "G4VIS_BUILD_RAYTRACERX_DRIVER G4VIS_USE_RAYTRACERX Allows for interactive ray-tracing " \
                "graphics through X11. The X11 package is required. [n]": "",
                "G4VIS_BUILD_VRML_DRIVER G4VIS_USE_VRML These driver generate VRML files, which describe " \
                "3D scenes to be visualized with a proper VRML viewer. [n]": "",
                "G4LIB_BUILD_GDML Setting this variable will enable building of the GDML plugin module " \
                "embedded in Geant4 for detector description persistency. It requires your system to have " \
                "the XercesC library and headers installed. Do you want to set this variable? [n]": "",
                "G4LIB_BUILD_G3TOG4 The utility module 'g3tog4' will be built by setting this variable. " \
                "NOTE: it requires a valid FORTRAN compiler to be installed on your system and the " \
                "'cernlib' command in the path, in order to build the ancillary tools! Do you want to " \
                "build 'g3tog4' ? [n]": "",
                "G4LIB_BUILD_ZLIB Do you want to activate compression for output files generated by the " \
                "HepRep visualization driver? [n]": "y",
                "G4ANALYSIS_USE Activates the configuration setup for allowing plugins to analysis tools " \
                "based on AIDA (Astract Interfaces for Data Analysis). In order to use AIDA features and " \
                "compliant analysis tools, the proper environment for these tools will have to be set " \
                "(see documentation for the specific analysis tools). [n]": "",
                "Press [Enter] to start installation or use a shell escape to edit config.sh: ": "",
                # extra questions and answers for version 9.2.p03
                "Directory %s doesn't exist. Use that name anyway? [n]" % dst: "y",
                "Specify the path where the Geant4 data libraries PhotonEvaporation%s " \
                "RadioactiveDecay%s G4EMLOW%s G4NDL%s G4ABLA%s are " \
                "installed. For now, a flat directory structure is assumed, and this can be customized " \
                "at the next step if needed. [%s/data]" % (self.cfg['PhotonEvaporationVersion'],
                                                           self.cfg['G4RadioactiveDecayVersion'],
                                                           self.cfg['G4EMLOWVersion'],
                                                           self.cfg['G4NDLVersion'],
                                                           self.cfg['G4ABLAVersion'],
                                                           pwd
                                                           ): "%s/data" % dst,
                "Please enter 1) Another path to search in 2) 'f' to force the use of the path " \
                "you entered previously (the data libraries are not needed to build Geant4, but " \
                "are needed to run applications later). 3) 'c' to customize the data paths, e.g. " \
                "if you have the data libraries installed in different locations. [f]": "",
                "G4UI_BUILD_QT_SESSION G4UI_USE_QT Setting these variables will enable the building " \
                "of the G4 Qt based user interface module and the use of this module in your " \
                "applications respectively. The Qt3 or Qt4 headers, libraries and moc application are " \
                "required to enable the building of this module. Do you want to enable build and use of " \
                "this module? [n]": "",
                # extra questions and answers for version 9.4.po1
                "What is the path to the Geant4 source tree? [%s]" % pwd: "",
                "Where should Geant4 be installed? [%s]" % pwd: dst,
                "Do you want to install all Geant4 headers in one directory? [n]": "y",
                "Do you want to build shared libraries? [y]": "",
                "Do you want to build static libraries too? [n]": "",
                "Do you want to build global libraries? [y]": "",
                "Do you want to build granular libraries as well? [n]": "",
                "Do you want to build libraries with debugging information? [n]": "",
                "Specify the path where the Geant4 data libraries are installed: [%s/data]" % pwd: "%s/data" % dst,
                "How many parallel jobs should make launch? [1]": "%s" % self.cfg['parallel'],
                "Please enter 1) Another path to search in 2) 'f' to force the use of the path you entered " \
                "previously (the data libraries are NOT needed to build Geant4, but are needed to run " \
                "applications later). 3) 'c' to customize the data paths, e.g. if you have the data " \
                "libraries installed in different locations. [f]": "",
                "Enable building of User Interface (UI) modules? [y]": "",
                "Enable building of the XAW (X11 Athena Widget set) UI module? [n]": "",
                "Enable building of the X11-Motif (Xm) UI module? [n]": "",
                "Enable building of the Qt UI module? [n]": "",
                "Enable building of visualization drivers? [y]": "n",
                "Enable the Geometry Description Markup Language (GDML) module? [n]": "",
                "Enable build of the g3tog4 utility module? [n]": "",
                "Enable internal zlib compression for HepRep visualization? [n] ": "",
            }

            self.noqanda = [
                r"Compiling\s+.*?\s+\.\.\.",
                r"Making\s+dependency\s+for\s+file\s+.*?\s+\.\.\.",
                r"Making\s+libname\.map\s+starter\s+file\s+\.\.\.",
                r"Making\s+libname\.map\s+\.\.\.",
                r"Reading\s+library\s+get_name\s+map\s+file\s*\.\.\.",
                r"Reading\s+dependency\s+files\s*\.\.\.",
                r"Creating\s+shared\s+library\s+.*?\s+\.\.\.",
            ]

            run_cmd_qa(cmd, self.qanda, self.noqanda, log_all=True, simple=True)

            # determining self.g4system
            try:
                scriptdirbase = os.path.join(pwd, '.config', 'bin')
                filelist = os.listdir(scriptdirbase)
            except OSError as err:
                raise EasyBuildError("Failed to determine self.g4system: %s", err)

            if len(filelist) != 1:
                raise EasyBuildError("Exactly one directory is expected in %s; found back: %s", scriptdirbase, filelist)
            else:
                self.g4system = filelist[0]

            self.scriptdir = os.path.join(scriptdirbase, self.g4system)
            if not os.path.isdir(self.scriptdir):
                raise EasyBuildError("Something went wrong. Dir: %s doesn't exist.", self.scriptdir)
            self.log.info("The directory containing several important scripts to be copied was found: %s",
                          self.scriptdir)

            # copying config.sh to pwd
            self.log.info("copying config.sh to %s", pwd)
            copy_file(os.path.join(self.scriptdir, 'config.sh'), pwd)

            # creating several scripts containing environment variables
            cmd = "%s/Configure -S -f config.sh -D g4conf=%s -D abssrc=%s" % (pwd, self.scriptdir, pwd)
            run_cmd(cmd, log_all=True, simple=True)

    def build_step(self):
        """Build Geant4."""
        if LooseVersion(self.version) >= LooseVersion("9.4"):
            super(EB_Geant4, self).build_step()
        else:
            pwd = self.cfg['start_dir']
            cmd = "%s/Configure -build" % pwd
            run_cmd_qa(cmd, self.qanda, no_qa=self.noqanda, log_all=True, simple=True)

    def install_step(self):
        """Install Geant4."""

        if LooseVersion(self.version) >= LooseVersion("9.4"):
            super(EB_Geant4, self).install_step()
            # '10.01.p03' -> '10.1.3'
            shortver = self.version.replace('.0', '.').replace('.p0', '.')
            self.datadst = os.path.join(self.installdir, 'share', '%s-%s' % (self.name, shortver), 'data')

            if LooseVersion(self.version) < LooseVersion("9.5"):
                version_parts = self.version.split('.')
                name = 'geant4-%s' % '.'.join(version_parts[:2] + [version_parts[-1][-1]])
                for (source, target) in [("%s.sh" % name, "env.sh"), ("%s.csh" % name, "env.csh")]:
                    source = os.path.join(self.installdir, 'share', name, 'config', source)
                    target = os.path.join(self.installdir, target)
                    try:
                        os.symlink(source, target)
                    except OSError as err:
                        raise EasyBuildError("Failed to symlink %s to %s: %s", source, target, err)
        else:
            pwd = self.cfg['start_dir']

            try:
                datasrc = os.path.join(pwd, '..')
                self.datadst = os.path.join(self.installdir, 'data')
                os.mkdir(self.datadst)
            except OSError as err:
                raise EasyBuildError("Failed to create data destination dir %s: %s", self.datadst, err)

            datalist = [
                'G4ABLA%s' % self.cfg['G4ABLAVersion'],
                'G4EMLOW%s' % self.cfg['G4EMLOWVersion'],
                'G4NDL%s' % self.cfg['G4NDLVersion'],
                'PhotonEvaporation%s' % self.cfg['PhotonEvaporationVersion'],
                'RadioactiveDecay%s' % self.cfg['G4RadioactiveDecayVersion'],
            ]
            try:
                for dat in datalist:
                    self.log.info("Copying %s to %s" % (dat, self.datadst))
                    shutil.copytree(os.path.join(datasrc, dat), os.path.join(self.datadst, dat))
            except IOError as err:
                raise EasyBuildError("Something went wrong during data copying (%s) to %s: %s", dat, self.datadst, err)

            try:
                for fil in ['config', 'environments', 'examples']:
                    self.log.info("Copying %s to %s" % (fil, self.installdir))
                    if not os.path.exists(os.path.join(pwd, fil)):
                        raise EasyBuildError("No such file or directory: %s", fil)
                    if os.path.isdir(os.path.join(pwd, fil)):
                        shutil.copytree(os.path.join(pwd, fil), os.path.join(self.installdir, fil))
                    elif os.path.isfile(os.path.join(pwd, fil)):
                        shutil.copy2(os.path.join(pwd, fil), os.path.join(self.installdir, fil))
            except IOError as err:
                raise EasyBuildError("Something went wrong during copying of %s to %s: %s", fil, self.installdir, err)

            try:
                for fil in ['config.sh', 'env.sh', 'env.csh']:
                    self.log.info("Copying %s to %s" % (fil, self.installdir))
                    if not os.path.exists(os.path.join(self.scriptdir, fil)):
                        raise EasyBuildError("No such file or directory: %s", fil)
                    if os.path.isdir(os.path.join(self.scriptdir, fil)):
                        shutil.copytree(os.path.join(self.scriptdir, fil), os.path.join(self.installdir, fil))
                    elif os.path.isfile(os.path.join(self.scriptdir, fil)):
                        shutil.copy2(os.path.join(self.scriptdir, fil), os.path.join(self.installdir, fil))
            except IOError as err:
                raise EasyBuildError("Something went wrong during copying of (%s) to %s: %s", fil, self.installdir, err)

            cmd = "%(pwd)s/Configure -f %(pwd)s/config.sh -d -install" % {'pwd': pwd}
            run_cmd(cmd, log_all=True, simple=True)

            mpiuidir = os.path.join(self.installdir, "examples/extended/parallel/MPI/mpi_interface")
            os.chdir(mpiuidir)

            # tweak config file as needed
            G4MPItxt = read_file('G4MPI.gmk')

            root_re = re.compile(r"(.*G4MPIROOT\s+=\s+).*", re.MULTILINE)
            cxx_re = re.compile(r"(.*CXX\s+:=\s+).*", re.MULTILINE)
            cppflags_re = re.compile(r"(.*CPPFLAGS\s+\+=\s+.*)", re.MULTILINE)

            G4MPItxt = root_re.sub(r"\1%s/intel64" % get_software_root('IMPI'), G4MPItxt)
            G4MPItxt = cxx_re.sub(r"\1mpicxx -cxx=icpc", G4MPItxt)
            G4MPItxt = cppflags_re.sub(r"\1 -I$(G4INCLUDE) -I%s)/include" % get_software_root('CLHEP'), G4MPItxt)

            self.log.debug("contents of G4MPI.gmk: %s" % G4MPItxt)

            shutil.copyfile("G4MPI.gmk", "G4MPI.gmk.ORIG")
            write_file('G4MPI.gmk', G4MPItxt)

            # make sure the required environment variables are there
            env.setvar("G4INSTALL", self.installdir)
            env.setvar("G4SYSTEM", self.g4system)
            env.setvar("G4LIB", "%s/lib/geant4/" % self.installdir)
            env.setvar("G4INCLUDE", "%s/include/geant4/" % self.installdir)

            run_cmd("make", log_all=True, simple=True)
            run_cmd("make includes", log_all=True, simple=True)

    def make_module_extra(self):
        """Define Geant4-specific environment variables in module file."""
        g4version = '.'.join(self.version.split('.')[:2])

        txt = super(EB_Geant4, self).make_module_extra()
        txt += self.module_generator.set_environment('G4INSTALL', self.installdir)
        # no longer needed in > 9.5, but leave it there for now.
        txt += self.module_generator.set_environment('G4VERSION', g4version)

        incdir = os.path.join(self.installdir, 'include')
        libdir = os.path.join(self.installdir, 'lib')
        if LooseVersion(self.version) >= LooseVersion("9.5"):
            txt += self.module_generator.set_environment('G4INCLUDE', os.path.join(incdir, 'Geant4'))
            txt += self.module_generator.set_environment('G4LIB', os.path.join(self.installdir, 'lib64', 'Geant4'))
        elif LooseVersion(self.version) >= LooseVersion("9.4"):
            txt += self.module_generator.set_environment('G4INCLUDE', os.path.join(incdir, 'geant4'))
            txt += self.module_generator.set_environment('G4LIB', libdir)
        else:
            txt += self.module_generator.set_environment('G4INCLUDE', os.path.join(incdir, 'geant4'))
            txt += self.module_generator.set_environment('G4LIB', os.path.join(libdir, 'geant4'))
            txt += self.module_generator.set_environment('G4SYSTEM', self.g4system)
            if self.cfg['G4ABLAVersion']:
                g4abladata = os.path.join(self.datadst, 'G4ABLA%s' % self.cfg['G4ABLAVersion'])
                txt += self.module_generator.set_environment('G4ABLADATA', g4abladata)

        if self.cfg['PhotonEvaporationVersion']:
            g4levelgammadata = os.path.join(self.datadst, 'PhotonEvaporation%s' % self.cfg['PhotonEvaporationVersion'])
            txt += self.module_generator.set_environment('G4LEVELGAMMADATA', g4levelgammadata)

        if self.cfg['G4RadioactiveDecayVersion']:
            g4radioactivedata = os.path.join(self.datadst, 'RadioactiveDecay%s' % self.cfg['G4RadioactiveDecayVersion'])
            txt += self.module_generator.set_environment('G4RADIOACTIVEDATA', g4radioactivedata)

        if self.cfg['G4EMLOWVersion']:
            g4ledata = os.path.join(self.datadst, 'G4EMLOW%s' % self.cfg['G4EMLOWVersion'])
            txt += self.module_generator.set_environment('G4LEDATA', g4ledata)

        if self.cfg['G4NDLVersion']:
            g4neutronhpdata = os.path.join(self.datadst, 'G4NDL%s' % self.cfg['G4NDLVersion'])
            txt += self.module_generator.set_environment('G4NEUTRONHPDATA', g4neutronhpdata)

        return txt

    def sanity_check_step(self):
        """
        Custom sanity check for Geant4
        """
        bin_files = ["bin/geant4-config"]
        if LooseVersion(self.version) >= LooseVersion("9.5"):
            bin_files.extend(["bin/geant4.sh", "bin/geant4.csh"])
            lib_files = ["lib64/libG4%s.so" % x for x in ['analysis', 'event', 'GMocren', 'materials',
                                                          'persistency', 'readout', 'Tree', 'VRML']]
            include_dir = 'include/Geant4'
        else:
            # paths for v9.4, untested for prior versions
            lib_files = ["lib/libG4%s.so" % x for x in ['event', 'GMocren', 'materials', 'persistency',
                                                        'readout', 'Tree', 'VRML']]
            include_dir = 'include/geant4'

        custom_paths = {
            'files': bin_files + lib_files,
            'dirs': [include_dir],
        }

        super(EB_Geant4, self).sanity_check_step(custom_paths)



# -*- coding: utf-8 -*-
##
# Copyright 2009-2023 Ghent University
#
# This file is part of EasyBuild,
# originally created by the HPC team of Ghent University (http://ugent.be/hpc/en),
# with support of Ghent University (http://ugent.be/hpc),
# the Flemish Supercomputer Centre (VSC) (https://vscentrum.be/nl/en),
# Flemish Research Foundation (FWO) (http://www.fwo.be/en)
# and the Department of Economy, Science and Innovation (EWI) (http://www.ewi-vlaanderen.be/en).
#
# https://github.com/easybuilders/easybuild
#
# EasyBuild is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation v2.
#
# EasyBuild is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with EasyBuild.  If not, see <http://www.gnu.org/licenses/>.
##
"""
EasyBuild support for installing Gurobi, implemented as an easyblock

@author: Bob Dröge (University of Groningen)
@author: Samuel Moors (Vrije Universiteit Brussel)
"""
import os

from easybuild.easyblocks.generic.pythonpackage import det_pylibdir
from easybuild.easyblocks.generic.tarball import Tarball
from easybuild.framework.easyconfig import CUSTOM
from easybuild.tools.build_log import EasyBuildError
from easybuild.tools.filetools import copy_file
from easybuild.tools.modules import get_software_root
from easybuild.tools.run import run_cmd


class EB_Gurobi(Tarball):
    """Support for installing linux64 version of Gurobi."""

    @staticmethod
    def extra_options(extra_vars=None):
        """Define extra options for Gurobi"""
        extra = {
            'copy_license_file': [True, "Copy license_file to installdir", CUSTOM],
        }
        return Tarball.extra_options(extra_vars=extra)

    def __init__(self, *args, **kwargs):
        """Easyblock constructor, define custom class variables specific to Gurobi."""
        super(EB_Gurobi, self).__init__(*args, **kwargs)

        self.license_file = self.cfg['license_file']

        if self.cfg['copy_license_file']:
            self.license_file = os.path.join(self.installdir, 'gurobi.lic')

    def install_step(self):
        """Install Gurobi and license file."""

        # make sure license file is available
        if self.cfg['license_file'] is None or not os.path.exists(self.cfg['license_file']):
            raise EasyBuildError("No existing license file specified: %s", self.cfg['license_file'])

        super(EB_Gurobi, self).install_step()

        if self.cfg['copy_license_file']:
            copy_file(self.cfg['license_file'], self.license_file)

        if get_software_root('Python'):
            run_cmd("python setup.py install --prefix=%s" % self.installdir)

    def sanity_check_step(self):
        """Custom sanity check for Gurobi."""
        custom_paths = {
            'files': ['bin/%s' % f for f in ['grbprobe', 'grbtune', 'gurobi_cl', 'gurobi.sh']],
            'dirs': ['matlab'],
        }

        custom_commands = [
            "gurobi_cl --help",
            'test -f $GRB_LICENSE_FILE',
        ]

        if get_software_root('Python'):
            custom_commands.append("python -c 'import gurobipy'")

        super(EB_Gurobi, self).sanity_check_step(custom_commands=custom_commands, custom_paths=custom_paths)

    def make_module_extra(self):
        """Custom extra module file entries for Gurobi."""
        txt = super(EB_Gurobi, self).make_module_extra()
        txt += self.module_generator.set_environment('GUROBI_HOME', self.installdir)
        txt += self.module_generator.set_environment('GRB_LICENSE_FILE', self.license_file)

        if get_software_root('Python'):
            txt += self.module_generator.prepend_paths('PYTHONPATH', det_pylibdir())

        txt += self.module_generator.prepend_paths('MATLABPATH', 'matlab')

        return txt



##
# Copyright 2009-2023 Ghent University
#
# This file is part of EasyBuild,
# originally created by the HPC team of Ghent University (http://ugent.be/hpc/en),
# with support of Ghent University (http://ugent.be/hpc),
# the Flemish Supercomputer Centre (VSC) (https://www.vscentrum.be),
# Flemish Research Foundation (FWO) (http://www.fwo.be/en)
# and the Department of Economy, Science and Innovation (EWI) (http://www.ewi-vlaanderen.be/en).
#
# https://github.com/easybuilders/easybuild
#
# EasyBuild is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation v2.
#
# EasyBuild is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with EasyBuild.  If not, see <http://www.gnu.org/licenses/>.
##
"""
EasyBuild support for pybind11, implemented as an easyblock

@author: Alexander Grund (TU Dresden)
"""
import os
from easybuild.easyblocks.generic.cmakemake import CMakeMake
from easybuild.easyblocks.generic.pythonpackage import PythonPackage
from easybuild.easyblocks.generic.cmakepythonpackage import CMakePythonPackage
import easybuild.tools.environment as env
from easybuild.tools.build_log import EasyBuildError
from easybuild.tools.filetools import change_dir
from easybuild.tools.run import run_cmd
from easybuild.tools.modules import get_software_root


class EB_pybind11(CMakePythonPackage):
    """Build PyBind11 for consumption with python packages and CMake

    PyBind11 can be consumed by CMake projects using `find_package` and by
    Python packages using `import pybind11`
    Hence we need to install PyBind11 twice: Once with CMake and once with pip
    """
    @staticmethod
    def extra_options(extra_vars=None):
        """Easyconfig parameters specific to PyBind11: Set defaults"""
        extra_vars = PythonPackage.extra_options(extra_vars=extra_vars)
        extra_vars = CMakeMake.extra_options(extra_vars=extra_vars)
        extra_vars['use_pip'][0] = True
        extra_vars['sanity_pip_check'][0] = True
        extra_vars['download_dep_fail'][0] = True
        return extra_vars

    def configure_step(self):
        """Avoid that a system Python is picked up when a Python module is loaded"""

        # make sure right 'python' command is used for installing pybind11
        python_root = get_software_root('Python')
        python_exe_opt = '-DPYTHON_EXECUTABLE='
        if python_root and python_exe_opt not in self.cfg['configopts']:
            configopt = python_exe_opt + os.path.join(python_root, 'bin', 'python')
            self.log.info("Adding %s to configopts since it is not specified yet", configopt)
            self.cfg.update('configopts', configopt)

        super(EB_pybind11, self).configure_step()

    def test_step(self):
        """Run pybind11 tests"""
        # run tests unless explicitly disabled
        if self.cfg['runtest'] is not False:
            self.cfg['runtest'] = 'check'
        super(EB_pybind11, self).test_step()

    def install_step(self):
        """Install with cmake install and pip install"""
        build_dir = change_dir(self.cfg['start_dir'])
        PythonPackage.install_step(self)

        # Reset installopts (set by PythonPackage)
        self.cfg['installopts'] = ''
        change_dir(build_dir)
        CMakeMake.install_step(self)

    def sanity_check_step(self):
        """
        Custom sanity check for Python packages
        """
        # don't add user site directory to sys.path (equivalent to python -s)
        env.setvar('PYTHONNOUSERSITE', '1', verbose=False)
        # Get python includes
        if not self.is_extension:
            # only load fake module for stand-alone installations (not for extensions),
            # since for extension the necessary modules should already be loaded at this point
            fake_mod_data = self.load_fake_module(purge=True)
        cmd = "%s -c 'import pybind11; print(pybind11.get_include())'" % self.python_cmd
        out, ec = run_cmd(cmd, simple=False)
        if ec:
            raise EasyBuildError("Failed to get pybind11 includes!")
        python_include = out.strip()
        if not self.is_extension:
            self.clean_up_fake_module(fake_mod_data)

        # Check for CMake config and includes
        custom_paths = {
            'files': ['share/cmake/pybind11/pybind11Config.cmake'],
            'dirs': ['include/pybind11', os.path.join(python_include, 'pybind11')],
        }
        # Check for Python module
        return PythonPackage.sanity_check_step(self, custom_paths=custom_paths)



##
# Copyright 2009-2023 Ghent University
#
# This file is part of EasyBuild,
# originally created by the HPC team of Ghent University (http://ugent.be/hpc/en),
# with support of Ghent University (http://ugent.be/hpc),
# the Flemish Supercomputer Centre (VSC) (https://www.vscentrum.be),
# Flemish Research Foundation (FWO) (http://www.fwo.be/en)
# and the Department of Economy, Science and Innovation (EWI) (http://www.ewi-vlaanderen.be/en).
#
# https://github.com/easybuilders/easybuild
#
# EasyBuild is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation v2.
#
# EasyBuild is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with EasyBuild.  If not, see <http://www.gnu.org/licenses/>.
##
"""
EasyBuild support for building and installing QScintilla, implemented as an easyblock

author: Kenneth Hoste (HPC-UGent)
author: Maxime Boissonneault (Compute Canada)
"""
import os
from distutils.version import LooseVersion

from easybuild.easyblocks.generic.configuremake import ConfigureMake
from easybuild.easyblocks.generic.pythonpackage import det_pylibdir
from easybuild.tools.build_log import EasyBuildError
from easybuild.tools.filetools import apply_regex_substitutions, mkdir, symlink, write_file, find_glob_pattern
from easybuild.tools.modules import get_software_root, get_software_version
from easybuild.tools.run import run_cmd
from easybuild.tools.systemtools import get_shared_lib_ext


class EB_QScintilla(ConfigureMake):
    """Support for building/installing QScintilla."""

    def prepare_step(self, *args, **kwargs):
        """Prepare build environment for building & installing QScintilla."""

        super(EB_QScintilla, self).prepare_step(*args, **kwargs)

        pyqt5 = get_software_root('PyQt5')
        pyqt = get_software_root('PyQt')
        qt5 = get_software_root('Qt5')
        if pyqt5:
            self.pyqt_root = pyqt5
            self.pyqt_pkg_name = "PyQt5"
        elif pyqt:
            self.pyqt_root = pyqt
            self.pyqt_pkg_name = "PyQt4"
        elif qt5:
            self.pyqt_root = qt5
            self.pyqt_pkg_name = "PyQt5"
        else:
            raise EasyBuildError("Failed to determine PyQt(5) installation prefix. Missing PyQt(5) dependency?")

        # PyQt5 is supported by QScintilla from version 2.11, otherwise there are some additional hacks/patches needed
        if LooseVersion(self.version) < LooseVersion('2.11') and pyqt5:
            raise EasyBuildError("PyQt5 is supported by QScintilla in version 2.11 and greater.")

    def configure_step(self):
        """Custom configuration procedure for QScintilla."""

        srcdir = os.path.join(self.cfg['start_dir'], 'Qt4Qt5')
        try:
            os.chdir(srcdir)
        except OSError as err:
            raise EasyBuildError("Failed to change to %s: %s", srcdir, err)

        # replace template values for install locations in qscintilla.pro configuration file
        regex_subs = [
            (r'\$\$\[QT_HOST_DATA\]', os.path.join(self.installdir, 'data')),
            (r'\$\$\[QT_INSTALL_DATA\]', os.path.join(self.installdir, 'data')),
            (r'\$\$\[QT_INSTALL_HEADERS\]', os.path.join(self.installdir, 'include')),
            (r'\$\$\[QT_INSTALL_LIBS\]', os.path.join(self.installdir, 'lib')),
            (r'\$\$\[QT_INSTALL_TRANSLATIONS\]', os.path.join(self.installdir, 'trans')),
        ]
        apply_regex_substitutions('qscintilla.pro', regex_subs)

        run_cmd("qmake qscintilla.pro")

    def build_step(self):
        """Custom build procedure for QScintilla."""

        # make sure that $CXXFLAGS is being passed down
        self.cfg.update('buildopts', r'CXXFLAGS="$CXXFLAGS \$(DEFINES)"')

        super(EB_QScintilla, self).build_step()

    def install_step(self):
        """Custom install procedure for QScintilla."""

        super(EB_QScintilla, self).install_step()

        # also install Python bindings if Python is included as a dependency
        python = get_software_root('Python')
        if python:
            pydir = os.path.join(self.cfg['start_dir'], 'Python')
            try:
                os.chdir(pydir)
            except OSError as err:
                raise EasyBuildError("Failed to change to %s: %s", pydir, err)

            # apparently this directory has to be there
            qsci_sipdir = os.path.join(self.installdir, 'share', 'sip', self.pyqt_pkg_name)
            mkdir(qsci_sipdir, parents=True)

            pylibdir = os.path.join(det_pylibdir(), self.pyqt_pkg_name)
            pyshortver = '.'.join(get_software_version('Python').split('.')[:2])

            sip_incdir = find_glob_pattern(os.path.join(self.pyqt_root, 'include', 'python%s*' % pyshortver), False)
            # depending on PyQt5 versions and how it was installed, the sip directory could be in various places
            # test them and figure out the first one that matches
            pyqt_sip_subdir = [os.path.join('share', 'python%s*' % pyshortver, 'site-packages', 'sip',
                                            self.pyqt_pkg_name),
                               os.path.join('share', 'sip', self.pyqt_pkg_name),
                               os.path.join('share', 'sip'),
                               os.path.join('lib', 'python%s*' % pyshortver, 'site-packages', self.pyqt_pkg_name,
                                            'bindings')
                               ]
            pyqt_sipdir_options = [os.path.join(self.pyqt_root, subdir) for subdir in pyqt_sip_subdir]
            for pyqt_sipdir_option in pyqt_sipdir_options:
                pyqt_sipdir = find_glob_pattern(pyqt_sipdir_option, False)
                if pyqt_sipdir:
                    break

            if not pyqt_sipdir:
                raise EasyBuildError("Failed to find PyQt5 sip directory")

            cfgopts = [
                '--destdir %s' % os.path.join(self.installdir, pylibdir),
                '--qsci-sipdir %s' % qsci_sipdir,
                '--qsci-incdir %s' % os.path.join(self.installdir, 'include'),
                '--qsci-libdir %s' % os.path.join(self.installdir, 'lib'),
                '--pyqt-sipdir %s' % pyqt_sipdir,
                '--apidir %s' % os.path.join(self.installdir, 'qsci', 'api', 'python'),
                '--no-stubs',
            ]
            if sip_incdir:
                cfgopts += ['--sip-incdir %s' % sip_incdir]

            if LooseVersion(self.version) >= LooseVersion('2.10.7'):
                cfgopts.append('--no-dist-info')

            # This flag was added in version 2.11
            if LooseVersion(self.version) >= LooseVersion('2.11'):
                cfgopts.append("--pyqt=%s" % self.pyqt_pkg_name)

            run_cmd("python configure.py %s" % ' '.join(cfgopts))

            super(EB_QScintilla, self).build_step()
            super(EB_QScintilla, self).install_step()

            target_dir = os.path.join(self.installdir, pylibdir)
            pyqt_pylibdir = os.path.join(self.pyqt_root, pylibdir)
            try:
                os.chdir(target_dir)
                for entry in [x for x in os.listdir(pyqt_pylibdir) if not x.startswith('__init__.py')]:
                    symlink(os.path.join(pyqt_pylibdir, entry), os.path.join(target_dir, entry))
            except OSError as err:
                raise EasyBuildError("Failed to symlink PyQt Python bindings in %s: %s", target_dir, err)

            # also requires empty __init__.py file to ensure Python modules can be imported from this location
            write_file(os.path.join(target_dir, '__init__.py'), '')

    def sanity_check_step(self):
        """Custom sanity check for QScintilla."""

        if LooseVersion(self.version) >= LooseVersion('2.10'):
            if self.pyqt_pkg_name == 'PyQt5':
                qsci_lib = 'libqscintilla2_qt5'
            else:
                qsci_lib = 'libqscintilla2_qt4'
        else:
            qsci_lib = 'libqscintilla2'

        custom_paths = {
            'files': [os.path.join('lib', qsci_lib + '.' + get_shared_lib_ext())],
            'dirs': ['data', os.path.join('include', 'Qsci'), 'trans'],
        }
        # also install Python bindings if Python is included as a dependency
        python = get_software_root('Python')

        custom_commands = []
        if python:
            custom_paths['dirs'].extend([
                os.path.join(det_pylibdir(), self.pyqt_pkg_name),
                os.path.join('qsci', 'api', 'python'),
                os.path.join('share', 'sip', self.pyqt_pkg_name),
            ])
            custom_commands.append("python -c 'import %s.Qsci'" % self.pyqt_pkg_name)

        super(EB_QScintilla, self).sanity_check_step(custom_paths=custom_paths, custom_commands=custom_commands)

    def make_module_extra(self):
        """Custom extra module file entries for QScintilla."""
        txt = super(EB_QScintilla, self).make_module_extra()
        python = get_software_root('Python')
        if python:
            if self.cfg['multi_deps'] and 'Python' in self.cfg['multi_deps']:
                txt += self.module_generator.prepend_paths('EBPYTHONPREFIXES', '')
            else:
                txt += self.module_generator.prepend_paths('PYTHONPATH', [det_pylibdir()])
        return txt



##
# Copyright 2009-2023 Ghent University
#
# This file is part of EasyBuild,
# originally created by the HPC team of Ghent University (http://ugent.be/hpc/en),
# with support of Ghent University (http://ugent.be/hpc),
# the Flemish Supercomputer Centre (VSC) (https://www.vscentrum.be),
# Flemish Research Foundation (FWO) (http://www.fwo.be/en)
# and the Department of Economy, Science and Innovation (EWI) (http://www.ewi-vlaanderen.be/en).
#
# https://github.com/easybuilders/easybuild
#
# EasyBuild is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation v2.
#
# EasyBuild is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with EasyBuild.  If not, see <http://www.gnu.org/licenses/>.
##
"""
EasyBuild support for installing the Intel Fortran compiler suite, implemented as an easyblock

@author: Stijn De Weirdt (Ghent University)
@author: Dries Verdegem (Ghent University)
@author: Kenneth Hoste (Ghent University)
@author: Pieter De Baets (Ghent University)
@author: Jens Timmerman (Ghent University)
@author: Ward Poelmans (Ghent University)
"""

import os
from distutils.version import LooseVersion

from easybuild.easyblocks.generic.intelbase import IntelBase
from easybuild.easyblocks.icc import EB_icc  # @UnresolvedImport
from easybuild.tools.systemtools import get_shared_lib_ext


class EB_ifort(EB_icc, IntelBase):
    """
    Class that can be used to install ifort
    - tested with 11.1.046
    -- will fail for all older versions (due to newer silent installer)
    """

    def sanity_check_step(self):
        """Custom sanity check paths for ifort."""
        shlib_ext = get_shared_lib_ext()

        binprefix = 'bin/intel64'
        libprefix = 'lib/intel64'
        if LooseVersion(self.version) >= LooseVersion('2011'):
            if LooseVersion(self.version) <= LooseVersion('2011.3.174'):
                binprefix = 'bin'
            elif LooseVersion(self.version) >= LooseVersion('2013_sp1'):
                binprefix = 'bin'
            else:
                libprefix = 'compiler/lib/intel64'

        bins = ['ifort']
        if LooseVersion(self.version) < LooseVersion('2013'):
            # idb is not shipped with ifort anymore in 2013.x versions (it is with icc though)
            bins.append('idb')

        libs = ['lib%s' % lib for lib in ['ifcore.a', 'ifcore.%s' % shlib_ext, 'iomp5.a', 'iomp5.%s' % shlib_ext]]
        custom_paths = {
            'files': [os.path.join(binprefix, x) for x in bins] + [os.path.join(libprefix, lib) for lib in libs],
            'dirs': [],
        }

        # make very sure that expected 'compilers_and_libraries_<VERSION>/linux' subdir is there for recent versions,
        # since we rely on it being there in make_module_req_guess
        if self.comp_libs_subdir:
            custom_paths['dirs'].append(self.comp_libs_subdir)

        custom_commands = ["which ifort"]

        IntelBase.sanity_check_step(self, custom_paths=custom_paths, custom_commands=custom_commands)

    def make_module_req_guess(self):
        """
        Additional paths to consider for prepend-paths statements in module file
        """
        guesses = super(EB_ifort, self).make_module_req_guess()
        if LooseVersion(self.version) >= LooseVersion('2016'):
            # This enables the creation of fortran 2008 bindings in MPI
            guesses['CPATH'].append('include')

        return guesses



##
# Copyright 2013-2023 Ghent University
#
# This file is part of EasyBuild,
# originally created by the HPC team of Ghent University (http://ugent.be/hpc/en),
# with support of Ghent University (http://ugent.be/hpc),
# the Flemish Supercomputer Centre (VSC) (https://www.vscentrum.be),
# Flemish Research Foundation (FWO) (http://www.fwo.be/en)
# and the Department of Economy, Science and Innovation (EWI) (http://www.ewi-vlaanderen.be/en).
#
# https://github.com/easybuilders/easybuild
#
# EasyBuild is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation v2.
#
# EasyBuild is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with EasyBuild.  If not, see <http://www.gnu.org/licenses/>.
##
"""
EasyBuild support for building and installing OpenIFS, implemented as an easyblock

@author: Kenneth Hoste (Ghent University)
"""
import os
import shutil

import easybuild.tools.environment as env
import easybuild.tools.toolchain as toolchain
from easybuild.framework.easyblock import EasyBlock
from easybuild.tools.build_log import EasyBuildError
from easybuild.tools.modules import get_software_root
from easybuild.tools.run import run_cmd


class EB_OpenIFS(EasyBlock):
    """Support for building/installing OpenIFS."""

    def configure_step(self):
        """Custom configuration procedure for OpenIFS."""
        # make sure use of MPI is enabled
        if not self.toolchain.options.get('usempi', None):
            raise EasyBuildError("Use of MPI should be enabled, set 'usempi' toolchain option to 'True'.")

        # configure build via OIFS_* environment variables
        env.setvar('OIFS_ARCH', 'x86_64')
        env.setvar('OIFS_BUILD', 'opt')
        if self.toolchain.comp_family() == toolchain.GCC:
            env.setvar('OIFS_COMP', 'gnu')
        elif self.toolchain.comp_family() == toolchain.INTELCOMP:
            env.setvar('OIFS_COMP', 'intel')
        else:
            raise EasyBuildError("Unknown compiler used, don't know how to set $OIFS_COMP.")

        # give EasyBuild control over compiler options
        env.setvar('OIFS_CC', os.getenv('CC'))
        env.setvar('OIFS_CFLAGS', os.getenv('CFLAGS'))
        env.setvar('OIFS_FC', os.getenv('F90'))
        env.setvar('OIFS_FFLAGS', os.getenv('F90FLAGS'))

        # set location of dependencies
        grib_api_root = get_software_root('grib_api')
        if grib_api_root:
            env.setvar('OIFS_GRIB_API_DIR', grib_api_root)
        else:
            raise EasyBuildError("grib_api module not loaded")

        env.setvar('OIFS_LAPACK_LIB', "-L%s %s" % (os.getenv('LAPACK_LIB_DIR'), os.getenv('LIBLAPACK')))

    def build_step(self):
        """Custom build procedure for OpenIFS."""
        try:
            os.chdir('make')
        except OSError as err:
            raise EasyBuildError("Failed to move to 'make' dir: %s", err)

        # enable parallel build
        par = self.cfg['parallel']
        cmd = "fcm make -v -j %s -f fcmcfg/oifs.cfg" % par
        run_cmd(cmd, log_all=True, simple=True, log_ok=True)

    def install_step(self):
        """Custom install procedure for OpenIFS: copy bin and include files."""
        try:
            srcdir = os.path.join(self.cfg['start_dir'], 'make', os.getenv('OIFS_BUILD'), 'oifs')
            bindir = os.path.join(self.installdir, 'bin')
            os.makedirs(bindir)
            shutil.copy2(os.path.join(srcdir, 'bin', 'master.exe'), bindir)
            shutil.copytree(os.path.join(srcdir, 'include'), os.path.join(self.installdir, 'include'))
        except OSError as err:
            raise EasyBuildError("Failed to install OpenIFS: %s", err)

    def sanity_check_step(self):
        """Custom sanity check for OpenIFS."""
        custom_paths = {
            'files': ['bin/master.exe'],
            'dirs': ['include'],
        }

        super(EB_OpenIFS, self).sanity_check_step(custom_paths=custom_paths)



##
# Copyright 2021-2023 Vrije Universiteit Brussel
#
# This file is part of EasyBuild,
# originally created by the HPC team of Ghent University (http://ugent.be/hpc/en),
# with support of Ghent University (http://ugent.be/hpc),
# the Flemish Supercomputer Centre (VSC) (https://www.vscentrum.be),
# Flemish Research Foundation (FWO) (http://www.fwo.be/en)
# and the Department of Economy, Science and Innovation (EWI) (http://www.ewi-vlaanderen.be/en).
#
# https://github.com/easybuilders/easybuild
#
# EasyBuild is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation v2.
#
# EasyBuild is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with EasyBuild.  If not, see <http://www.gnu.org/licenses/>.
##
"""
EasyBuild support for installing a wrapper module file for OpenSSL

@author: Alex Domingo (Vrije Universiteit Brussel)
"""
import os
import re

from distutils.version import LooseVersion

from easybuild.easyblocks.generic.bundle import Bundle
from easybuild.framework.easyconfig import CUSTOM
from easybuild.tools.build_log import EasyBuildError, print_warning
from easybuild.tools.filetools import change_dir, expand_glob_paths, mkdir, read_file, symlink, which, write_file
from easybuild.tools.py2vs3 import string_type
from easybuild.tools.run import run_cmd
from easybuild.tools.systemtools import DARWIN, LINUX, get_os_type, get_shared_lib_ext, find_library_path


class EB_OpenSSL_wrapper(Bundle):
    """
    Find path to installation files of OpenSSL in the host system. Checks in
    order: library files defined in 'openssl_libs', engines libraries, header
    files and executables. Any missing component will trigger an installation
    from source of the fallback component.
    Libraries are located by soname using the major and minor subversions of
    the wrapper version. The full version of the wrapper or the option
    'minimum_openssl_version' determine the minimum required version of OpenSSL
    in the system. The wrapper checks for version strings in the library files
    and the opensslv.h header.
    If OpenSSL in host systems fulfills the version requirements, wrap it by
    symlinking all installation files. Otherwise fall back to the bundled
    component.
    """

    @staticmethod
    def extra_options(extra_vars=None):
        """Easyconfig parameters specific to OpenSSL wrapper"""
        extra_vars = Bundle.extra_options(extra_vars=extra_vars)
        extra_vars.update({
            'wrap_system_openssl': [True, 'Detect and wrap OpenSSL installation in host system', CUSTOM],
            'minimum_openssl_version': [None, 'Minimum version of OpenSSL required in host system', CUSTOM],
        })
        return extra_vars

    def __init__(self, *args, **kwargs):
        """Locate the installation files of OpenSSL in the host system"""
        super(EB_OpenSSL_wrapper, self).__init__(*args, **kwargs)

        # Wrapper should have at least a major minor version numbers
        try:
            subversions = self.version.split('.')
            self.majmin_version = '%s.%s' % (subversions[0], subversions[1])
        except (AttributeError, IndexError):
            err_msg = "Wrapper OpenSSL version does not have any subversion: %s"
            raise EasyBuildError(err_msg, self.version)

        # Set minimum OpenSSL version
        min_openssl_version = self.cfg.get('minimum_openssl_version')

        if not min_openssl_version:
            min_openssl_version = self.version
        elif not isinstance(min_openssl_version, string_type):
            min_openssl_version = str(min_openssl_version)

        # Minimum OpenSSL version can only increase depth of wrapper version
        if min_openssl_version.startswith(self.version):
            self.log.debug("Requiring minimum OpenSSL version: %s", min_openssl_version)
        else:
            err_msg = "Requested minimum OpenSSL version '%s' does not fit in wrapper easyconfig version '%s'"
            raise EasyBuildError(err_msg, min_openssl_version, self.version)

        # Regex pattern to find version strings in OpenSSL libraries and headers
        full_version_regex = re.compile(r'[0-9]+\.[0-9]+\.[0-9]+[a-z]?')
        openssl_version_regex = re.compile(r'OpenSSL\s+([0-9]+\.[0-9]+(\.[0-9]+[a-z]?)*)', re.M)

        # Libraries packaged in OpenSSL
        openssl_libs = ['libssl', 'libcrypto']
        # list of relevant library extensions per system and version of OpenSSL
        # the first item should be the extension of an installation from source,
        # it will be used in the sanity checks of the component
        openssl_libext = {
            '1.0': {
                LINUX: ('so.1.0.0', 'so.10'),
                DARWIN: ('1.0.dylib', ),
            },
            '1.1': {
                LINUX: ('so.1.1', ),
                DARWIN: ('1.1.dylib', ),
            },
            '3.0': {
                LINUX: ('so.3', ),
                DARWIN: ('3.dylib', ),
            },
        }

        os_type = get_os_type()
        if self.majmin_version in openssl_libext and os_type in openssl_libext[self.majmin_version]:
            # generate matrix of versioned .so filenames
            system_versioned_libs = [
                ['%s.%s' % (lib, ext) for lib in openssl_libs]
                for ext in openssl_libext[self.majmin_version][os_type]
            ]
            self.log.info("Matrix of version library names: %s", system_versioned_libs)
        else:
            err_msg = "Don't know name of OpenSSL system library for version %s and OS type %s"
            raise EasyBuildError(err_msg, self.majmin_version, os_type)

        # by default target the first option of each OpenSSL library,
        # which corresponds to installation from source
        self.target_ssl_libs = system_versioned_libs[0]
        self.log.info("Target OpenSSL libraries: %s", self.target_ssl_libs)

        # folders containing engines libraries
        openssl_engines = {
            '1.0': 'engines',
            '1.1': 'engines-1.1',
            '3.0': 'engines-3',
        }
        self.target_ssl_engine = openssl_engines[self.majmin_version]

        # Paths to system libraries and headers of OpenSSL
        self.system_ssl = {
            'bin': None,
            'engines': None,
            'include': None,
            'libs': [],
        }

        # early return when we're not wrapping the system OpenSSL installation
        if not self.cfg.get('wrap_system_openssl'):
            self.log.info("Not wrapping system OpenSSL installation by user request")
            return

        # Check the system libraries of OpenSSL
        # Find library file and compare its version string
        for idx, solibs in enumerate(system_versioned_libs):
            for solib in solibs:
                system_solib = find_library_path(solib)
                if system_solib:
                    openssl_version = 0
                    # get version of system library filename
                    try:
                        openssl_version = full_version_regex.search(os.path.realpath(system_solib)).group(0)
                    except AttributeError:
                        # filename lacks the full version, fallback to version strings within the library
                        solib_strings = read_file(system_solib, mode="rb").decode('utf-8', 'replace')
                        try:
                            openssl_version = openssl_version_regex.search(solib_strings).group(1)
                        except AttributeError:
                            dbg_msg = "Could not detect the full version of system OpenSSL library: %s"
                            self.log.debug(dbg_msg, system_solib)
                    # check that system version fulfills requirements
                    if LooseVersion(openssl_version) >= LooseVersion(min_openssl_version):
                        dbg_msg = "System OpenSSL library '%s' with version %s fulfills requested version %s"
                        self.log.debug(dbg_msg, system_solib, openssl_version, min_openssl_version)
                        self.system_ssl['libs'].append(system_solib)
                    else:
                        dbg_msg = "System OpenSSL library '%s' with version %s is older than requested version %s"
                        self.log.debug(dbg_msg, system_solib, openssl_version, min_openssl_version)
                else:
                    # one of the OpenSSL libraries is missing, switch to next group of versioned libs
                    self.system_ssl['libs'] = []
                    break

            if len(self.system_ssl['libs']) == len(openssl_libs):
                # keep these libraries as possible targets for this installation
                target_system_ssl_libs = system_versioned_libs[idx]
                break

        if len(self.system_ssl['libs']) == len(openssl_libs):
            self.system_ssl['version'] = openssl_version
            info_msg = "Found OpenSSL library version %s in host system: %s"
            self.log.info(info_msg, self.system_ssl['version'], os.path.dirname(self.system_ssl['libs'][0]))
        else:
            self.log.info("OpenSSL library not found in host system, falling back to OpenSSL in EasyBuild")
            return

        # Directory with engine libraries
        lib_dir = os.path.dirname(self.system_ssl['libs'][0])
        lib_engines_dir = [
            os.path.join(lib_dir, 'openssl', self.target_ssl_engine),
            os.path.join(lib_dir, self.target_ssl_engine),
        ]

        for engines_path in lib_engines_dir:
            if os.path.isdir(engines_path):
                self.system_ssl['engines'] = engines_path
                self.log.debug("Found OpenSSL engines in: %s", self.system_ssl['engines'])
                break

        if not self.system_ssl['engines']:
            self.log.info("OpenSSL engines not found in host system, falling back to OpenSSL in EasyBuild")
            return

        # Check system include paths for OpenSSL headers
        cmd = "LC_ALL=C gcc -E -Wp,-v -xc /dev/null"
        (out, ec) = run_cmd(cmd, log_all=True, simple=False, trace=False)

        sys_include_dirs = []
        for match in re.finditer(r'^\s(/[^\0\n]*)+', out, re.MULTILINE):
            sys_include_dirs.extend(match.groups())
        self.log.debug("Found the following include directories in host system: %s", ', '.join(sys_include_dirs))

        # headers are located in 'include/openssl' by default
        ssl_include_subdirs = ['openssl']
        if self.majmin_version == '1.1':
            # but version 1.1 can be installed in 'include/openssl11/openssl' as well, for example in CentOS 7
            # prefer 'include/openssl' as long as the version of headers matches
            ssl_include_subdirs.append(os.path.join('openssl11', self.name.lower()))

        ssl_include_dirs = [os.path.join(incd, subd) for incd in sys_include_dirs for subd in ssl_include_subdirs]
        ssl_include_dirs = [include for include in ssl_include_dirs if os.path.isdir(include)]

        # find location of header files for this version of the OpenSSL libraries
        for include_dir in ssl_include_dirs:
            opensslv_path = os.path.join(include_dir, 'opensslv.h')
            self.log.debug("Checking OpenSSL version in %s...", opensslv_path)
            if os.path.exists(opensslv_path):
                # check version reported by opensslv.h
                opensslv = read_file(opensslv_path)
                try:
                    header_version = openssl_version_regex.search(opensslv).group(1)
                except AttributeError:
                    err_msg = "System OpenSSL header '%s' does not contain any recognizable version string"
                    raise EasyBuildError(err_msg, opensslv_path)

                if header_version == self.system_ssl['version']:
                    self.system_ssl['include'] = include_dir
                    info_msg = "Found OpenSSL headers v%s in host system: %s"
                    self.log.info(info_msg, header_version, self.system_ssl['include'])
                    break
                else:
                    dbg_msg = "System OpenSSL header version '%s' doesn not match library version '%s'"
                    self.log.debug(dbg_msg, header_version, self.system_ssl['version'])
            else:
                self.log.info("System OpenSSL header file %s not found", opensslv_path)

        if not self.system_ssl['include']:
            err_msg = ("OpenSSL v%s headers not found in host system, but libraries for v%s are present. "
                       "Install the development package of OpenSSL for your system or force building OpenSSL from "
                       "source in EasyBuild by setting 'wrap_system_openssl = False' in the OpenSSL easyconfig.")
            raise EasyBuildError(err_msg, self.version, self.system_ssl['version'])

        # Check system OpenSSL binary
        if self.majmin_version == '1.1':
            # prefer 'openssl11' over 'openssl' with v1.1
            self.system_ssl['bin'] = which('openssl11')

        if not self.system_ssl['bin']:
            self.system_ssl['bin'] = which('openssl')

        if self.system_ssl['bin']:
            self.log.info("System OpenSSL binary found: %s", self.system_ssl['bin'])
        else:
            self.log.info("System OpenSSL binary not found!")
            return

        # system OpenSSL is fine, change target libraries to the ones found in it
        self.target_ssl_libs = target_system_ssl_libs
        self.log.info("Target system OpenSSL libraries: %s", self.target_ssl_libs)

    def fetch_step(self, *args, **kwargs):
        """Fetch sources if OpenSSL component is needed"""
        if not all(self.system_ssl[key] for key in ('bin', 'engines', 'include', 'libs')):
            super(EB_OpenSSL_wrapper, self).fetch_step(*args, **kwargs)

    def extract_step(self):
        """Extract sources if OpenSSL component is needed"""
        if not all(self.system_ssl[key] for key in ('bin', 'engines', 'include', 'libs')):
            super(EB_OpenSSL_wrapper, self).extract_step()

    def install_step(self):
        """Symlink target OpenSSL installation"""
        if all(self.system_ssl[key] for key in ('bin', 'engines', 'include', 'libs')):
            # note: symlink to individual files, not directories,
            # since directory symlinks get resolved easily...

            # link OpenSSL libraries in system
            lib64_dir = os.path.join(self.installdir, 'lib64')
            lib64_engines_dir = os.path.join(lib64_dir, os.path.basename(self.system_ssl['engines']))
            mkdir(lib64_engines_dir, parents=True)

            # link existing known libraries
            for libso in self.system_ssl['libs']:
                symlink(libso, os.path.join(lib64_dir, os.path.basename(libso)))

            # link engines library files
            engine_lib_pattern = [os.path.join(self.system_ssl['engines'], '*')]
            for engine_lib in expand_glob_paths(engine_lib_pattern):
                symlink(engine_lib, os.path.join(lib64_engines_dir, os.path.basename(engine_lib)))

            # relative symlink for unversioned libraries
            cwd = change_dir(lib64_dir)
            for libso in self.system_ssl['libs']:
                libso = os.path.basename(libso)
                unversioned_lib = '%s.%s' % (libso.split('.')[0], get_shared_lib_ext())
                symlink(libso, unversioned_lib, use_abspath_source=False)
            change_dir(cwd)

            # link OpenSSL headers in system
            include_dir = os.path.join(self.installdir, 'include', self.name.lower())
            mkdir(include_dir, parents=True)
            include_pattern = [os.path.join(self.system_ssl['include'], '*')]
            for header_file in expand_glob_paths(include_pattern):
                symlink(header_file, os.path.join(include_dir, os.path.basename(header_file)))

            # link OpenSSL binary in system
            bin_dir = os.path.join(self.installdir, 'bin')
            mkdir(bin_dir)
            symlink(self.system_ssl['bin'], os.path.join(bin_dir, self.name.lower()))

            # install pkg-config files
            self.install_pc_files()

        elif self.cfg.get('wrap_system_openssl'):
            # install OpenSSL component due to lack of OpenSSL in host system
            print_warning("Not all OpenSSL components found in host system, falling back to OpenSSL in EasyBuild!")
            super(EB_OpenSSL_wrapper, self).install_step()
        else:
            # install OpenSSL component by user request
            warn_msg = "Installing OpenSSL from source in EasyBuild by user request ('wrap_system_openssl=%s')"
            print_warning(warn_msg, self.cfg.get('wrap_system_openssl'))
            super(EB_OpenSSL_wrapper, self).install_step()

    def sanity_check_step(self):
        """Custom sanity check for OpenSSL wrapper."""
        shlib_ext = get_shared_lib_ext()

        ssl_libs = ['%s.%s' % (libso.split('.')[0], shlib_ext) for libso in self.target_ssl_libs]
        ssl_libs.extend(self.target_ssl_libs)

        ssl_files = [os.path.join('bin', self.name.lower())]
        ssl_files.extend(os.path.join('lib', libso) for libso in ssl_libs)

        ssl_dirs = [
            os.path.join('include', self.name.lower()),
            os.path.join('lib', self.target_ssl_engine),
            os.path.join('lib', 'pkgconfig'),
        ]

        custom_paths = {
            'files': ssl_files,
            'dirs': ssl_dirs,
        }

        custom_commands = [
            # make sure that version mentioned in output of 'openssl version' matches version we are using
            "ssl_ver=$(openssl version); [ ${ssl_ver:8:3} == '%s' ]" % self.majmin_version,
            "echo | openssl s_client -connect github.com:443 -verify 9 | grep 'Verify return code: 0 (ok)'",
        ]

        super(Bundle, self).sanity_check_step(custom_paths=custom_paths, custom_commands=custom_commands)

    def install_pc_files(self):
        """Install pkg-config files for the wrapper"""

        PC_TEMPLATE = """prefix=%(root)s
exec_prefix=${prefix}
libdir=${exec_prefix}/lib64
includedir=${prefix}/include
%(enginesdir)s

Name: %(name)s
Description: %(description)s
Version: %(version)s
%(requires)s
%(libs)s
%(cflags)s
"""
        openssl_components = {
            'libcrypto': {
                'name': 'OpenSSL-libcrypto',
                'description': 'OpenSSL cryptography library',
                'enginesdir': self.target_ssl_engine,
            },
            'libssl': {
                'name': 'OpenSSL-libssl',
                'description': 'Secure Sockets Layer and cryptography libraries',
                'enginesdir': None,
            },
            'openssl': {
                'name': 'OpenSSL',
                'description': 'Secure Sockets Layer and cryptography libraries and tools',
                'enginesdir': None,
            },
        }

        pc_install_dir = os.path.join(self.installdir, 'lib64', 'pkgconfig')
        mkdir(pc_install_dir)

        for pc_comp in openssl_components:
            pc_file = openssl_components[pc_comp]

            pc_file['root'] = self.installdir
            pc_file['version'] = self.system_ssl['version']

            # component name in system pkg-config
            pc_name = pc_comp
            if self.majmin_version == '1.1':
                # check suffixed names with v1.1
                pc_name_suffix = pc_name + '11'
                pc_exists_cmd = "pkg-config --exists %s" % pc_name_suffix
                if run_cmd(pc_exists_cmd, simple=True, log_ok=False, log_all=False):
                    self.log.info("%s exists", pc_name_suffix)
                    pc_name = pc_name_suffix

            # get requires from pkg-config
            pc_file['requires'] = []
            for require_type in ['Requires', 'Requires.private']:
                require_print = require_type.lower().replace('.', '-')
                pc_print_cmd = "pkg-config --print-%s %s" % (require_print, pc_name)
                out, _ = run_cmd(pc_print_cmd, simple=False, log_ok=False)
                self.log.info("Output of '%s': %s", pc_print_cmd, out)

                if out:
                    requires = out
                    # use unsuffixed names for components provided by this wrapper
                    for wrap_comp in openssl_components:
                        requires = re.sub(r'^%s[0-9]+$' % wrap_comp, wrap_comp, requires, flags=re.M)
                    # format requires
                    requires = requires.rstrip().splitlines()
                    pc_file['requires'].append("%s: %s" % (require_type, ' '.join(requires)))

            pc_file['requires'] = '\n'.join(pc_file['requires'])

            if pc_comp.startswith('lib'):
                # add libs and cflags for library components
                c_lib_name = pc_comp[3:]
                pc_file['libs'] = "Libs: -L${libdir} -l%s" % c_lib_name
                pc_file['cflags'] = "Cflags: -I${includedir}"
                # infer private libs through pkg-config
                pc_libs_cmd = "pkg-config --libs %s" % pc_name
                out, _ = run_cmd(pc_libs_cmd, simple=False, log_ok=False)
                self.log.info("Output of '%s': %s", pc_libs_cmd, out)
                linker_libs = out

                pc_libs_static_cmd = "pkg-config --libs --static %s" % pc_name
                out, _ = run_cmd(pc_libs_static_cmd, simple=False, log_ok=False)
                self.log.info("Output of '%s': %s", pc_libs_static_cmd, out)

                libs_priv = "%s " % out.rstrip()
                for flag in linker_libs.rstrip().split(' '):
                    libs_priv = libs_priv.replace("%s " % flag, '')
                pc_file['libs'] += "\nLibs.private: %s" % libs_priv
            else:
                pc_file['libs'], pc_file['cflags'] = '', ''

            # format enginesdir
            if pc_file['enginesdir']:
                pc_file['enginesdir'] = 'enginesdir=${libdir}/%s' % pc_file['enginesdir']
            else:
                pc_file['enginesdir'] = ''

            pc_path = os.path.join(pc_install_dir, '%s.pc' % pc_comp)
            write_file(pc_path, PC_TEMPLATE % pc_file)



##
# Copyright 2009-2023 Ghent University
#
# This file is part of EasyBuild,
# originally created by the HPC team of Ghent University (http://ugent.be/hpc/en),
# with support of Ghent University (http://ugent.be/hpc),
# the Flemish Supercomputer Centre (VSC) (https://www.vscentrum.be),
# Flemish Research Foundation (FWO) (http://www.fwo.be/en)
# and the Department of Economy, Science and Innovation (EWI) (http://www.ewi-vlaanderen.be/en).
#
# https://github.com/easybuilders/easybuild
#
# EasyBuild is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation v2.
#
# EasyBuild is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with EasyBuild.  If not, see <http://www.gnu.org/licenses/>.
##
"""
EasyBuild support for wxPython, implemented as an easyblock

@author: Balazs Hajgato (Vrije Universiteit Brussel)
@author: Kenneth Hoste (HPC-UGent)
@author: Maxime Boissonneault (Compute Canada, Calcul Quebec, Universite Laval)
"""
import glob
import os

from distutils.version import LooseVersion
from easybuild.easyblocks.generic.pythonpackage import PythonPackage, det_python_version
from easybuild.tools.filetools import change_dir, symlink
from easybuild.tools.modules import get_software_root
from easybuild.tools.run import run_cmd
from easybuild.tools.systemtools import get_shared_lib_ext


class EB_wxPython(PythonPackage):
    """Support for installing the wxPython Python package."""

    def build_step(self):
        """Custom build step for wxPython."""
        if LooseVersion(self.version) >= LooseVersion("4"):
            prebuild_opts = self.cfg['prebuildopts']
            script = 'build.py'
            self.wxflag = ''
            if get_software_root('wxWidgets'):
                self.wxflag = '--use_syswx'

            BUILD_CMD = "%(prebuild_opts)s %(pycmd)s %(script)s --prefix=%(prefix)s -v"
            base_cmd = BUILD_CMD % {
                'prebuild_opts': prebuild_opts,
                'pycmd': self.python_cmd,
                'script': script,
                'prefix': self.installdir,
            }

            # Do we need to build wxWidgets internally?
            if self.wxflag == '':
                cmd = base_cmd + " build_wx"
                run_cmd(cmd, log_all=True, simple=True)

            cmd = base_cmd + " %s build_py" % self.wxflag
            run_cmd(cmd, log_all=True, simple=True)

    def install_step(self):
        """Custom install procedure for wxPython."""
        # wxPython configure, build, and install with one script
        preinst_opts = self.cfg['preinstallopts']
        INSTALL_CMD = "%(preinst_opts)s %(pycmd)s %(script)s --prefix=%(prefix)s"
        if LooseVersion(self.version) >= LooseVersion("4"):
            script = 'build.py'
            cmd = INSTALL_CMD % {
                'preinst_opts': preinst_opts,
                'pycmd': self.python_cmd,
                'script': script,
                'prefix': self.installdir,
            }
            # install fails and attempts to install in the python module. building the wheel, and then installing it
            cmd = cmd + " %s -v bdist_wheel" % self.wxflag
            run_cmd(cmd, log_all=True, simple=True)

            # get whether it is 35, 36, 37, 38, etc.
            pyver = det_python_version(self.python_cmd)
            pyver = pyver[0] + pyver[2]

            cmd = "pip install --no-deps --prefix=%(prefix)s dist/wxPython-%(version)s-cp%(pyver)s*.whl" % {
                'prefix': self.installdir,
                'version': self.version,
                'pyver': pyver
            }
        else:
            script = os.path.join('wxPython', 'build-wxpython.py')
            cmd = INSTALL_CMD % {
                'preinst_opts': preinst_opts,
                'pycmd': self.python_cmd,
                'script': script,
                'prefix': self.installdir,
            }
            cmd = cmd + " --wxpy_installdir=%s --install" % self.installdir

        run_cmd(cmd, log_all=True, simple=True)

        # add symbolic links for libwx_*so.* files
        # (which are created automatically by 'build.py install', but not by 'pip install *.whl')
        if LooseVersion(self.version) >= LooseVersion("4"):
            wx_lib_dir = os.path.join(self.installdir, self.pylibdir, 'wx')
            cwd = change_dir(wx_lib_dir)
            lib_so_files = glob.glob('libwx*.so.*')
            for lib_so_file in lib_so_files:
                symlink(lib_so_file, lib_so_file[:-2])
            change_dir(cwd)

    def sanity_check_step(self):
        """Custom sanity check for wxPython."""
        majver = '.'.join(self.version.split('.')[:2])
        shlib_ext = get_shared_lib_ext()
        py_bins = ['crust', 'shell', 'wxrc']
        files = []
        dirs = []
        if LooseVersion(self.version) < LooseVersion("4"):
            files.extend([os.path.join('bin', 'wxrc')])
            dirs.extend(['include', 'share'])
            py_bins.extend(['alacarte', 'alamode', 'wrap'])
        elif LooseVersion(self.version) >= LooseVersion("4.1"):
            majver = '3.1'  # this is 3.1 in ver 4.1.x
            py_bins.extend(['slices', 'slicesshell'])
        elif LooseVersion(self.version) >= LooseVersion("4.0"):
            majver = '3.0'  # for some reason this is still 3.0 in ver 4.0.x
            py_bins.extend(['slices', 'slicesshell'])

        files.extend([os.path.join('bin', 'py%s' % x) for x in py_bins])
        dirs.extend([self.pylibdir])

        if LooseVersion(self.version) < LooseVersion("4"):
            libfiles = ['lib%s-%s.%s' % (x, majver, shlib_ext) for x in ['wx_baseu', 'wx_gtk2u_core']]
            files.extend([os.path.join('lib', f) for f in libfiles])
        else:
            if not get_software_root('wxWidgets'):
                libfiles = ['lib%s-%s.%s' % (x, majver, shlib_ext) for x in ['wx_baseu', 'wx_gtk3u_core']]
                files.extend([os.path.join(self.pylibdir, 'wx', f) for f in libfiles])

        custom_paths = {
            'files': files,
            'dirs': dirs,
        }

        # test using 'import wx' (i.e. don't use 'import wxPython')
        self.options['modulename'] = 'wx'

        if LooseVersion(self.version) < LooseVersion("4"):
            # also test importing wxversion
            custom_commands = [(self.python_cmd, '-c "import wxversion"')]
        else:
            # also test importing wx.lib.wxcairo
            custom_commands = [(self.python_cmd, '-c "import wx.lib.wxcairo"')]

        return super(EB_wxPython, self).sanity_check_step(custom_paths=custom_paths, custom_commands=custom_commands)

    def make_module_extra(self):
        """Custom update for $PYTHONPATH for wxPython."""
        txt = super(EB_wxPython, self).make_module_extra()

        if LooseVersion(self.version) < LooseVersion("4"):
            # make sure that correct subdir is also included to $PYTHONPATH
            majver = '.'.join(self.version.split('.')[:2])
            txt += self.module_generator.prepend_paths('PYTHONPATH', os.path.join(self.pylibdir, 'wx-%s-gtk2' % majver))

        return txt



##
# Copyright 2009-2023 Ghent University
#
# This file is part of EasyBuild,
# originally created by the HPC team of Ghent University (http://ugent.be/hpc/en),
# with support of Ghent University (http://ugent.be/hpc),
# the Flemish Supercomputer Centre (VSC) (https://www.vscentrum.be),
# Flemish Research Foundation (FWO) (http://www.fwo.be/en)
# and the Department of Economy, Science and Innovation (EWI) (http://www.ewi-vlaanderen.be/en).
#
# https://github.com/easybuilders/easybuild
#
# EasyBuild is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation v2.
#
# EasyBuild is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with EasyBuild.  If not, see <http://www.gnu.org/licenses/>.
##
"""
EasyBuild support for OpenSSL, implemented as an easyblock

@author: Kenneth Hoste (Ghent University)
@author: Jens Timmerman (Ghent University)
@author: Davide Vanzo (ACCRE - Vanderbilt University)
@author: Alex Domingo (Vrije Universiteit Brussel)
"""
import os
import re

from distutils.version import LooseVersion

from easybuild.easyblocks.generic.configuremake import ConfigureMake
from easybuild.framework.easyconfig import CUSTOM
from easybuild.tools.build_log import EasyBuildError, print_warning
from easybuild.tools.filetools import remove_dir, symlink
from easybuild.tools.run import run_cmd
from easybuild.tools.systemtools import get_shared_lib_ext

GENERIC_SSL_CERTS_DIR = "/etc/ssl/certs"


class EB_OpenSSL(ConfigureMake):
    """Support for building OpenSSL"""

    @staticmethod
    def extra_options(extra_vars=None):
        """Easyconfig parameters specific to OpenSSL"""
        extra_vars = ConfigureMake.extra_options(extra_vars=extra_vars)
        extra_vars.update({
            'ssl_certificates': [None, "Absolute path to 'certs' directory with the system SSL certificates", CUSTOM],
        })
        return extra_vars

    def __init__(self, *args, **kwargs):
        """Initialisation of custom class variables for OpenSSL"""
        super(EB_OpenSSL, self).__init__(*args, **kwargs)

        # path to SSL certificates
        ssl_certs_dir = self.cfg.get('ssl_certificates')

        if ssl_certs_dir is not None:
            # check option ssl_certificates
            ssl_certs_dir = os.path.normpath(ssl_certs_dir)
            if not os.path.isabs(ssl_certs_dir):
                raise EasyBuildError("ssl_certificates is not an absolute path: %s", ssl_certs_dir)
            if os.path.basename(ssl_certs_dir) != 'certs':
                raise EasyBuildError("ssl_certificates does not point to a 'certs' directory: %s", ssl_certs_dir)
            if not os.path.isdir(ssl_certs_dir):
                raise EasyBuildError("ssl_certificates 'certs' directory does not exist: %s", ssl_certs_dir)
        else:
            # set ssl_certs_dir from system OPENSSLDIR
            openssldir = ''
            openssldir_regex = re.compile(r'^OPENSSLDIR: "(.*)"$')
            openssldir_cmd = "openssl version -d"

            try:
                (out, _) = run_cmd(openssldir_cmd, log_all=True, simple=False, trace=False)
                openssldir = openssldir_regex.search(out).group(1)
            except EasyBuildError:
                self.log.info("OPENSSLDIR not found in system (openssl command failed), "
                              "continuing with generic OPENSSLDIR path...")
            except AttributeError:
                self.log.debug("OPENSSLDIR not found in system (openssl reported '%s'), "
                               "continuing with generic OPENSSLDIR path...", out)
            else:
                self.log.info("OPENSSLDIR determined from system openssl: %s", openssldir)

            if os.path.isdir(openssldir):
                ssl_certs_dir = os.path.join(openssldir, 'certs')
            elif os.path.isdir(GENERIC_SSL_CERTS_DIR):
                # fallback to generic OPENSSLDIR
                ssl_certs_dir = GENERIC_SSL_CERTS_DIR
                self.log.info("Falling back to generic SSL certificates directory: %s", ssl_certs_dir)
            else:
                self.log.info("Generic SSL certificates directory not found: %s", GENERIC_SSL_CERTS_DIR)

        self.ssl_certs_dir = ssl_certs_dir
        self.log.debug("SSL certificates directory: %s", self.ssl_certs_dir)

    def configure_step(self, cmd_prefix=''):
        """
        Configure step
        """

        cmd = "%s %s./config --prefix=%s threads shared %s" % (self.cfg['preconfigopts'], cmd_prefix,
                                                               self.installdir, self.cfg['configopts'])

        (out, _) = run_cmd(cmd, log_all=True, simple=False)

        return out

    def install_step(self):
        """Installation of OpenSSL and SSL certificates"""
        super(EB_OpenSSL, self).install_step()

        # SSL certificates
        # OPENSSLDIR is already populated by the installation of OpenSSL
        # try to symlink system certificates in the empty 'certs' directory
        ssl_dir = os.path.join(self.installdir, 'ssl')
        openssl_certs_dir = os.path.join(ssl_dir, 'certs')

        if self.ssl_certs_dir:
            remove_dir(openssl_certs_dir)
            symlink(self.ssl_certs_dir, openssl_certs_dir)

            # also symlink cert.pem file, if it exists
            # (required on CentOS 7, see https://github.com/easybuilders/easybuild-easyconfigs/issues/14058)
            cert_pem_path = os.path.join(os.path.dirname(self.ssl_certs_dir), 'cert.pem')
            if os.path.isfile(cert_pem_path):
                symlink(cert_pem_path, os.path.join(ssl_dir, os.path.basename(cert_pem_path)))
        else:
            print_warning("OpenSSL successfully installed without system SSL certificates. "
                          "Some packages might experience limited functionality.")

    def sanity_check_step(self):
        """Custom sanity check"""

        # basic paths
        custom_paths = {
            'files': ['bin/openssl'],
            'dirs': ['include', 'ssl'],
        }

        # add libraries
        lib_dir = 'lib'
        lib_sonames = ['libcrypto', 'libssl']
        shlib_ext = get_shared_lib_ext()
        lib_files = [os.path.join(lib_dir, '%s.%s') % (x, y) for x in lib_sonames for y in ['a', shlib_ext]]

        custom_paths['files'].extend(lib_files)

        # add engines
        engines_dir = 'engines'
        if LooseVersion(self.version) >= LooseVersion("1.1"):
            engines_dir = 'engines-1.1'

        custom_paths['dirs'].append(os.path.join(lib_dir, engines_dir))

        # add SSL certificates
        if self.ssl_certs_dir:
            custom_paths['dirs'].append('ssl/certs')

        super(EB_OpenSSL, self).sanity_check_step(custom_paths=custom_paths)



##
# Copyright 2009-2023 Ghent University
#
# This file is part of EasyBuild,
# originally created by the HPC team of Ghent University (http://ugent.be/hpc/en),
# with support of Ghent University (http://ugent.be/hpc),
# the Flemish Supercomputer Centre (VSC) (https://www.vscentrum.be),
# Flemish Research Foundation (FWO) (http://www.fwo.be/en)
# and the Department of Economy, Science and Innovation (EWI) (http://www.ewi-vlaanderen.be/en).
#
# https://github.com/easybuilders/easybuild
#
# EasyBuild is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation v2.
#
# EasyBuild is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with EasyBuild.  If not, see <http://www.gnu.org/licenses/>.
##
"""
EasyBuild support for building and installing Ferret, implemented as an easyblock

@author: Stijn De Weirdt (Ghent University)
@author: Dries Verdegem (Ghent University)
@author: Kenneth Hoste (Ghent University)
@author: Pieter De Baets (Ghent University)
@author: Jens Timmerman (Ghent University)
@author: George Fanourgakis (The Cyprus Institute)
@author: Samuel Moors (Vrije Universiteit Brussel (VUB))
"""


import os
from distutils.version import LooseVersion
import easybuild.tools.toolchain as toolchain
from easybuild.easyblocks.generic.configuremake import ConfigureMake
from easybuild.tools.build_log import EasyBuildError
from easybuild.tools.filetools import apply_regex_substitutions, change_dir, copy_file
from easybuild.tools.modules import get_software_root


class EB_Ferret(ConfigureMake):
    """Support for building/installing Ferret."""

    def configure_step(self):
        """Configure Ferret build."""

        buildtype = "x86_64-linux"
        if LooseVersion(self.version) < LooseVersion("7.3"):
            change_dir('FERRET')

        if LooseVersion(self.version) >= LooseVersion("7.5"):
            deps = ['HDF5', 'netCDF-Fortran']
        else:
            deps = ['HDF5', 'netCDF', 'Java']

        for name in deps:
            if not get_software_root(name):
                raise EasyBuildError("%s module not loaded?", name)

        if LooseVersion(self.version) >= LooseVersion("7.3"):
            copy_file('external_functions/ef_utility/site_specific.mk.in',
                      'external_functions/ef_utility/site_specific.mk')
            copy_file('site_specific.mk.in', 'site_specific.mk')
            fns = [
                "site_specific.mk",
                "external_functions/ef_utility/site_specific.mk",
            ]
        else:
            fns = ["site_specific.mk"]

        regex_subs = [
            (r"^BUILDTYPE\s*=.*", "BUILDTYPE = %s" % buildtype),
            (r"^INSTALL_FER_DIR =.*", "INSTALL_FER_DIR = %s" % self.installdir),
        ]

        if LooseVersion(self.version) >= LooseVersion("7.5"):
            regex_subs.append((r"^(HDF5_LIBDIR\s*)=.*", r"\1 = %s/lib" % get_software_root('HDF5')))
            regex_subs.append((r"^(NETCDF_LIBDIR\s*)=.*", r"\1 = %s/lib" % get_software_root('netCDF-Fortran')))
        else:
            for name in deps:
                regex_subs.append((r"^(%s.*DIR\s*)=.*" % name.upper(), r"\1 = %s" % get_software_root(name)))

        if LooseVersion(self.version) >= LooseVersion("7.3"):
            regex_subs.extend([
                (r"^DIR_PREFIX =.*", "DIR_PREFIX = %s" % self.cfg['start_dir']),
                (r"^FER_LOCAL_EXTFCNS = $(FER_DIR)", "FER_LOCAL_EXTFCNS = $(INSTALL_FER_DIR)/libs"),
            ])

        if LooseVersion(self.version) >= LooseVersion("7.5"):
            comp_vars = {
                'CC': 'CC',
                'FC': 'FC',
            }

            for key, value in comp_vars.items():
                regex_subs.append((r"^(%s\s*)=.*" % key, r"\1= %s " % os.getenv(value)))

            if self.toolchain.comp_family() == toolchain.INTELCOMP:
                regex_subs.append((r"^(\s*LD\s*)=.*", r"\1 = %s -nofor-main " % os.getenv("FC")))

        for fn in fns:
            apply_regex_substitutions(fn, regex_subs)

        comp_vars = {
            'CC': 'CC',
            'CFLAGS': 'CFLAGS',
            'CPPFLAGS': 'CPPFLAGS',
            'FC': 'F77',
        }

        gfort2ifort = {
            '-fno-second-underscore': ' ',
            '-fno-backslash': ' ',
            '-fdollar-ok': ' ',
            '-ffast-math': ' ',
            '-ffixed-line-length-132': '-132',
            '-fno-automatic': ' ',
            '-ffpe-trap=overflow': ' ',
            '-fimplicit-none': '-implicitnone',
            '-fdefault-real-8': '-r8',
            '-fdefault-double-8': ' ',
            '-Wl,-Bstatic -lgfortran -Wl,-Bdynamic': ' ',
            '-v --verbose -m64': ' ',
            '-export-dynamic': ' ',
            '-DG77_SIGNAL': ' ',
        }

        fn = 'xgks/CUSTOMIZE.%s' % buildtype

        regex_subs = [(r"^(FFLAGS\s*=').*-m64 (.*)", r"\1%s \2" % os.getenv('FFLAGS'))]
        for key, value in comp_vars.items():
            regex_subs.append((r"^(%s\s*)=.*" % key, r"\1='%s'" % os.getenv(value)))

        x11_root = get_software_root('X11')
        if x11_root:
            regex_subs.append((r"^(LD_X11\s*)=.*", r"\1='-L%s/lib -lX11'" % x11_root))
        else:
            regex_subs.append((r"^(LD_X11\s*)=.*", r"\1='-L/usr/lib64/X11 -lX11'"))

        if LooseVersion(self.version) >= LooseVersion("7.3") and self.toolchain.comp_family() == toolchain.INTELCOMP:
            regex_subs.extend(sorted(gfort2ifort.items()))

        apply_regex_substitutions(fn, regex_subs)

        comp_vars = {
            'CC': 'CC',
            'CXX': 'CXX',
            'F77': 'F77',
            'FC': 'F77',
        }

        if LooseVersion(self.version) >= LooseVersion("7.3"):
            fns = [
                'platform_specific.mk.%s' % buildtype,
                'external_functions/ef_utility/platform_specific.mk.%s' % buildtype,
            ]
        else:
            fns = [
                'fer/platform_specific_flags.mk.%s' % buildtype,
                'ppl/platform_specific_flags.mk.%s' % buildtype,
                'external_functions/ef_utility/platform_specific_flags.mk.%s' % buildtype,
            ]

        regex_subs = []
        for key, value in comp_vars.items():
            regex_subs.append((r"^(\s*%s\s*)=.*" % key, r"\1 = %s" % os.getenv(value)))

        if LooseVersion(self.version) >= LooseVersion("7.3"):
            regex_subs.extend([
                (r"^(\s*LDFLAGS\s*=).*", r"\1 -fPIC %s -lnetcdff -lnetcdf -lhdf5_hl -lhdf5" % os.getenv("LDFLAGS")),
                (r"^(\s*)CDFLIB", r"\1NONEED"),
            ])

        if self.toolchain.comp_family() == toolchain.INTELCOMP:
            regex_subs.append((r"^(\s*LD\s*)=.*", r"\1 = %s -nofor-main" % os.getenv("F77")))
            for x in ["CFLAGS", "FFLAGS"]:
                regex_subs.append((r"^(\s*%s\s*=\s*\$\(CPP_FLAGS\)).*\\" % x, r"\1 %s \\" % os.getenv(x)))
            if LooseVersion(self.version) >= LooseVersion("7.3"):
                for x in ["CFLAGS", "FFLAGS"]:
                    regex_subs.append((r"^(\s*%s\s*=).*-m64 (.*)" % x, r"\1%s \2" % os.getenv(x)))
                regex_subs.extend(sorted(gfort2ifort.items()))

                regex_subs.append((r"^(\s*MYDEFINES\s*=.*)\\", r"\1-DF90_SYSTEM_ERROR_CALLS \\"))

        for fn in fns:
            apply_regex_substitutions(fn, regex_subs)

        if LooseVersion(self.version) >= LooseVersion("7.3"):
            comp_vars = {
                'CC': 'CC',
                'LDFLAGS': 'LDFLAGS',
            }
            fn = 'gksm2ps/Makefile'

            regex_subs = [(r"^(\s*CFLAGS=\")-m64 (.*)", r"\1%s \2" % os.getenv('CFLAGS'))]
            for key, value in comp_vars.items():
                regex_subs.append((r"^(\s*%s)=.*" % key, r"\1='%s' \\" % os.getenv(value)))

            apply_regex_substitutions(fn, regex_subs)

    def sanity_check_step(self):
        """Custom sanity check for Ferret."""

        major_minor_version = '.'.join(self.version.split('.')[:2])
        custom_paths = {
            'files': ["bin/ferret_v%s" % major_minor_version],
            'dirs': [],
        }

        super(EB_Ferret, self).sanity_check_step(custom_paths=custom_paths)



# -*- coding: utf-8 -*-
##
# Copyright 2009-2023 Ghent University
#
# This file is part of EasyBuild,
# originally created by the HPC team of Ghent University (http://ugent.be/hpc/en),
# with support of Ghent University (http://ugent.be/hpc),
# the Flemish Supercomputer Centre (VSC) (https://www.vscentrum.be),
# Flemish Research Foundation (FWO) (http://www.fwo.be/en)
# and the Department of Economy, Science and Innovation (EWI) (http://www.ewi-vlaanderen.be/en).
#
# https://github.com/easybuilders/easybuild
#
# EasyBuild is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation v2.
#
# EasyBuild is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with EasyBuild.  If not, see <http://www.gnu.org/licenses/>.
##
"""
EasyBuild support for building and installing AOMP version of LLVM/Clang

@author: Jørgen Nordmoen (University of Oslo, USIT)
"""
import glob
import os

from distutils.version import LooseVersion
from easybuild.easyblocks.clang import DEFAULT_TARGETS_MAP as LLVM_ARCH_MAP
from easybuild.easyblocks.generic.bundle import Bundle
from easybuild.framework.easyblock import EasyBlock
from easybuild.framework.easyconfig import CUSTOM
from easybuild.tools.build_log import EasyBuildError, print_msg, print_warning
from easybuild.tools.modules import get_software_root
from easybuild.tools.systemtools import AARCH64, POWER, X86_64, get_cpu_architecture, get_shared_lib_ext


# Default AMD GPU architectures to build for
#
# AMD uses 'gfx' to identify the GPU, the first number identifies the generation, according to
# https://www.x.org/wiki/RadeonFeature/#index5h2 while the rest identifies the specific GPU.
# In the context of EasyBuild this identifier can be thought of as equivalent to the 'sm_<xx>'
# nomenclature of Nvidia.
DEFAULT_GFX_ARCHS = ['gfx900', 'gfx902', 'gfx906', 'gfx908', 'gfx90a', 'gfx1030', 'gfx1031']


class EB_Clang_minus_AOMP(Bundle):
    """
    Self-contained build of AOMP version of Clang
    """

    @staticmethod
    def extra_options():
        gfx_list_help_msg = "AMD GPU architectures to build for (if None, use defaults: %s)"
        extra_vars = {
            'gfx_list': [None, gfx_list_help_msg % ', '.join(DEFAULT_GFX_ARCHS), CUSTOM],
        }
        return Bundle.extra_options(extra_vars)

    def __init__(self, *args, **kwargs):
        """Easyblock constructor."""
        super(EB_Clang_minus_AOMP, self).__init__(*args, **kwargs)

        # List of LLVM target architectures to build for, extended in the 'prepare_step'
        self.target_archs = ['AMDGPU']

        # Mapping from known ROCm components to their configure method
        self.cfg_method = {
            'aomp-extras': self._configure_aomp_extras,
            'Clang-OpenMP': self._configure_omp,
            'llvm-project': self._configure_llvm,
            'ROCm-Device-Libs': self._configure_rocm_device_libs,
        }

        # Prepare configuration options that point to the expected Clang build
        self.llvm_compiler_flags = [
            "-DCMAKE_C_COMPILER=%s" % os.path.join(self.installdir, 'bin', 'clang'),
            "-DCMAKE_CXX_COMPILER=%s" % os.path.join(self.installdir, 'bin', 'clang++'),
            "-DLLVM_INSTALL_PREFIX=%s" % self.installdir,
            "-DLLVM_DIR=%s" % self.installdir,
        ]

        # Variables to be filled in the prepare step
        self.cuda_archs = []
        self.device_lib_path = None

        # Setup AMD GFX list to build for
        if self.cfg['gfx_list']:
            self.amd_gfx_archs = self.cfg['gfx_list']
        else:
            self.amd_gfx_archs = DEFAULT_GFX_ARCHS

    def prepare_step(self, *args, **kwargs):
        """
        Prepare build environment
        """
        super(EB_Clang_minus_AOMP, self).prepare_step(*args, **kwargs)

        # Detect CPU architecture and setup build targets for LLVM
        cpu_arch = get_cpu_architecture()
        if cpu_arch in LLVM_ARCH_MAP:
            self.target_archs.append(LLVM_ARCH_MAP[cpu_arch][0])
        else:
            raise EasyBuildError('Unknown CPU architecture for LLVM: %s', cpu_arch)

        # Set up target architectures for LLVM
        # If CUDA is loaded when building, build CUDA offload backend
        if get_software_root('CUDA'):
            self.target_archs.append('NVPTX')

            # a specific set of CUDA compute capabilities must be specified,
            # via --cuda-compute-capabilities EasyBuild configuration option or
            # via cuda_compute_capabilities easyconfig parameter
            cuda_cc = self.cfg.get_cuda_cc_template_value('cuda_compute_capabilities').split(',')
            if not cuda_cc:
                raise EasyBuildError("Can't build Clang-AOMP with CUDA support "
                                     "without specifying 'cuda-compute-capabilities'")
            self.cuda_archs = [cc.replace('.', '') for cc in cuda_cc]
            self.log.info("Building offload support for the following CUDA architectures: '%s'",
                          ' '.join(self.cuda_archs))

        self.log.info("Building LLVM for the following architectures: '%s'", ';'.join(self.target_archs))

        self.log.info("Building offload support for the following AMD architectures: '%s'",
                      ' '.join(self.amd_gfx_archs))

    def configure_step(self):
        """
        Go through each component and setup configuration for the later Bundle install step
        """
        super(EB_Clang_minus_AOMP, self).configure_step()

        # Ensure necessary libraries are downloaded and can be found
        device_lib_dir_pattern = os.path.join(self.builddir, 'ROCm-Device-Libs-*')
        hits = glob.glob(device_lib_dir_pattern)
        if len(hits) == 1:
            self.device_lib_path = hits[0]
        else:
            raise EasyBuildError("Could not find 'ROCm-Device-Libs' source directory in %s", self.builddir)

        num_comps = len(self.cfg['components'])
        for idx, comp in enumerate(self.comp_cfgs):
            name = comp['name']
            msg = "configuring bundle component %s %s (%d/%d)..." % (name, comp['version'], idx + 1, num_comps)
            print_msg(msg)
            if name in self.cfg_method:
                self.cfg_method[name](comp)
                self.log.info(msg)
            else:
                self.log.warn("Component %s has no configure method!" % name)

    def sanity_check_step(self):
        """
        Custom sanity check for ROCm
        """
        shlib_ext = get_shared_lib_ext()
        custom_paths = {
            'files': ['bin/clang', 'bin/lld', 'lib/libomp.%s' % shlib_ext,
                      'lib/libomptarget.rtl.amdgpu.%s' % shlib_ext, 'lib/libomptarget.%s' % shlib_ext],
            'dirs': ['amdgcn/bitcode', 'include/clang', 'include/lld', 'include/llvm'],
        }
        custom_commands = ['clang --help', 'clang++ --help']

        if LooseVersion(self.version) >= LooseVersion("5"):
            libs = ['aompextras']
        else:
            libs = ['aompextras', 'omptarget']
        libdevice = os.path.join('lib', 'libdevice')

        # Check that all AMD GFX libraries were built
        for gfx in self.amd_gfx_archs:
            if LooseVersion(self.version) < LooseVersion("5.2"):
                custom_paths['files'].extend([os.path.join(libdevice, 'lib%s-amdgcn-%s.bc' % (x, gfx)) for x in libs])
            if LooseVersion(self.version) >= LooseVersion("5"):
                custom_paths['files'].append(os.path.join('lib', 'libomptarget-amdgcn-%s.bc' % gfx))
                custom_paths['files'].append(os.path.join('lib', 'libomptarget-new-amdgpu-%s.bc' % gfx))

        if LooseVersion(self.version) >= LooseVersion("5"):
            custom_paths['files'].append(os.path.join('lib', 'libomptarget.rtl.amdgpu.%s' % shlib_ext))

        # Check that CPU target OpenMP offloading library was built
        arch = get_cpu_architecture()

        # Check architecture explicitly since Clang uses potentially different names
        if arch == X86_64:
            arch = 'x86_64'
        elif arch == POWER:
            arch = 'ppc64'
        elif arch == AARCH64:
            arch = 'aarch64'
        else:
            print_warning("Unknown CPU architecture (%s) for OpenMP offloading!" % arch)

        custom_paths['files'].append(os.path.join('lib', 'libomptarget.rtl.%s.%s' % (arch, shlib_ext)))

        # If CUDA offloading support was requested, check that correct omptarget was built
        if get_software_root('CUDA'):
            custom_paths['files'].append(os.path.join('lib', 'libomptarget.rtl.cuda.%s' % shlib_ext))

            for arch in self.cuda_archs:
                sm_arch = 'sm_%s' % arch
                if LooseVersion(self.version) >= LooseVersion("5"):
                    custom_paths['files'].append(os.path.join(libdevice, 'libm-nvptx-%s.bc' % sm_arch))
                    custom_paths['files'].append(os.path.join('lib', 'libomptarget-new-nvptx-%s.bc' % sm_arch))
                else:
                    custom_paths['files'].append(os.path.join('lib', 'libomptarget-nvptx-%s.bc' % sm_arch))

        # need to bypass sanity_check_step of Bundle, because it only loads the generated module
        # unless custom paths or commands are specified in the easyconfig
        EasyBlock.sanity_check_step(self, custom_paths=custom_paths, custom_commands=custom_commands)

    def _configure_llvm(self, component):
        """
        Setup configure options for building compiler_rt, Clang and lld
        """
        comp_dir = '%s-%s' % (component['name'], component['version'])
        component['srcdir'] = os.path.join(comp_dir, 'llvm')

        # Need to unset $CPATH to avoid that libunwind is pulled in via Mesa
        # dependency and interrupts building of LLVM
        component['prebuildopts'] = "unset CPATH && "

        # Setup configuration options for LLVM
        component['configopts'] = ' '.join([
            "-DLLVM_ENABLE_PROJECTS='clang;lld;compiler-rt'",
            "-DCLANG_DEFAULT_LINKER=lld",
            "-DGCC_INSTALL_PREFIX=$EBROOTGCC",
            "-DLLVM_ENABLE_ASSERTIONS=ON",
            "-DLLVM_ENABLE_BINDINGS=OFF",
            "-DLLVM_INCLUDE_BENCHMARKS=OFF",
            "-DLLVM_TARGETS_TO_BUILD='%s'" % ';'.join(self.target_archs),
        ])

    def _configure_rocm_device_libs(self, component):
        """
        Setup ROCm device libs such that it is built with the local LLVM build
        """
        component['configopts'] = ' '.join(self.llvm_compiler_flags + ['-DBUILD_HC_LIB=OFF'])

    def _configure_omp(self, component):
        """
        Setup OpenMP configuration options, OMP uses the LLVM source
        """
        llvm_include_dir = os.path.join(self.installdir, 'include', 'llvm')
        comp_dir = 'llvm-project-%s' % component['version']
        component['srcdir'] = os.path.join(comp_dir, 'openmp')

        component['preconfigopts'] = "export HSA_RUNTIME_PATH=%s && " % self.installdir

        component['configopts'] = ' '.join(self.llvm_compiler_flags + [
            "-DLIBOMPTARGET_AMDGCN_GFXLIST='%s'" % ';'.join(self.amd_gfx_archs),
            "-DLIBOMPTARGET_ENABLE_DEBUG=ON",
            "-DLIBOMPTARGET_LLVM_INCLUDE_DIRS=%s" % llvm_include_dir,
            "-DLIBOMP_COPY_EXPORTS=OFF",
            "-DLLVM_MAIN_INCLUDE_DIR=%s" % llvm_include_dir,
            "-DOPENMP_ENABLE_LIBOMPTARGET=ON",
            "-DOPENMP_ENABLE_LIBOMPTARGET_HSA=ON",
            "-DROCDL=%s" % self.device_lib_path,
            "-DROCM_DIR=%s" % self.installdir,
            "-DAOMP_STANDALONE_BUILD=1",
            '',
        ])

        if LooseVersion(self.version) >= LooseVersion("5.2"):
            component['configopts'] += ' '.join([
                "-DDEVICELIBS_ROOT=%s" % self.device_lib_path,
            ])

        if get_software_root('CUDA'):
            llvm_link = os.path.join(self.installdir, 'bin', 'llvm-link')
            cuda_path = os.path.join(self.installdir, 'bin', 'clang++')
            component['configopts'] += ' '.join([
                "-DLIBOMPTARGET_NVPTX_BC_LINKER=%s" % llvm_link,
                "-DLIBOMPTARGET_NVPTX_COMPUTE_CAPABILITIES='%s'" % ','.join(self.cuda_archs),
                "-DLIBOMPTARGET_NVPTX_CUDA_COMPILER=%s" % cuda_path,
                "-DLIBOMPTARGET_NVPTX_ENABLE_BCLIB=ON",
            ])

    def _configure_aomp_extras(self, component):
        """
        Setup AOMP extras compiler configurations
        """
        component['preconfigopts'] = "export AOMP=%s && " % self.installdir
        component['preconfigopts'] += "export GFXLIST='%s' && " % ';'.join(self.amd_gfx_archs)

        component['configopts'] = ' '.join(self.llvm_compiler_flags + [
            "-DAOMP_STANDALONE_BUILD=1",
            "-DROCDL=%s" % self.device_lib_path,
            "-DROCM_DIR=%s" % self.installdir,
            '',
        ])

        if LooseVersion(self.version) >= LooseVersion("5.2"):
            component['configopts'] += ' '.join([
                "-DAOMP_VERSION_STRING=%s" % self.version,
            ])



# #
# Copyright 2013-2023 Ghent University
#
# This file is part of EasyBuild,
# originally created by the HPC team of Ghent University (http://ugent.be/hpc/en),
# with support of Ghent University (http://ugent.be/hpc),
# the Flemish Supercomputer Centre (VSC) (https://www.vscentrum.be),
# Flemish Research Foundation (FWO) (http://www.fwo.be/en)
# and the Department of Economy, Science and Innovation (EWI) (http://www.ewi-vlaanderen.be/en).
#
# https://github.com/easybuilders/easybuild
#
# EasyBuild is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation v2.
#
# EasyBuild is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with EasyBuild.  If not, see <http://www.gnu.org/licenses/>.
# #
"""
EasyBuild support for installing Intel Inspector, implemented as an easyblock

@author: Kenneth Hoste (Ghent University)
@author: Damian Alvarez (Forschungzentrum Juelich GmbH)
"""
import os
from distutils.version import LooseVersion

from easybuild.easyblocks.generic.intelbase import IntelBase, ACTIVATION_NAME_2012, LICENSE_FILE_NAME_2012


class EB_Inspector(IntelBase):
    """
    Support for installing Intel Inspector
    """

    def __init__(self, *args, **kwargs):
        """Easyblock constructor; define class variables."""
        super(EB_Inspector, self).__init__(*args, **kwargs)

        # recent versions of Inspector are installed to a subdirectory
        self.subdir = ''
        loosever = LooseVersion(self.version)
        if loosever >= LooseVersion('2013_update7') and loosever < LooseVersion('2017'):
            self.subdir = 'inspector_xe'
        elif loosever >= LooseVersion('2017') and loosever < LooseVersion('2021'):
            self.subdir = 'inspector'
        elif loosever >= LooseVersion('2021'):
            self.subdir = os.path.join('inspector', 'latest')

    def make_installdir(self):
        """Do not create installation directory, install script handles that already."""
        super(EB_Inspector, self).make_installdir(dontcreate=True)

    def install_step(self):
        """
        Actual installation
        - create silent cfg file
        - execute command
        """
        silent_cfg_names_map = None

        if LooseVersion(self.version) <= LooseVersion('2013_update6'):
            silent_cfg_names_map = {
                'activation_name': ACTIVATION_NAME_2012,
                'license_file_name': LICENSE_FILE_NAME_2012,
            }

        super(EB_Inspector, self).install_step(silent_cfg_names_map=silent_cfg_names_map)

    def make_module_req_guess(self):
        """Find reasonable paths for Inspector"""
        return self.get_guesses_tools()

    def sanity_check_step(self):
        """Custom sanity check paths for Intel Inspector."""
        binaries = ['inspxe-cl', 'inspxe-feedback', 'inspxe-gui', 'inspxe-runmc', 'inspxe-runtc']
        custom_paths = self.get_custom_paths_tools(binaries)
        super(EB_Inspector, self).sanity_check_step(custom_paths=custom_paths)



##
# Copyright 2009-2023 Ghent University
#
# This file is part of EasyBuild,
# originally created by the HPC team of Ghent University (http://ugent.be/hpc/en),
# with support of Ghent University (http://ugent.be/hpc),
# the Flemish Supercomputer Centre (VSC) (https://www.vscentrum.be),
# Flemish Research Foundation (FWO) (http://www.fwo.be/en)
# and the Department of Economy, Science and Innovation (EWI) (http://www.ewi-vlaanderen.be/en).
#
# https://github.com/easybuilders/easybuild
#
# EasyBuild is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation v2.
#
# EasyBuild is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with EasyBuild.  If not, see <http://www.gnu.org/licenses/>.
##
"""
EasyBuild support for JAR files, implemented as an easyblock

@author: Stijn De Weirdt (Ghent University)
@author: Dries Verdegem (Ghent University)
@author: Kenneth Hoste (Ghent University)
@author: Pieter De Baets (Ghent University)
@author: Jens Timmerman (Ghent University)
"""
import shutil

from easybuild.easyblocks.generic.binary import Binary


class JAR(Binary):
    """Support for installing JAR files."""

    def install_step_xxx(self):
        """Custom installation for JAR files: just copy them to install path."""

        for srcfile in self.src:
            shutil.copy(srcfile['path'], self.installdir)

    def make_module_extra(self):
        """Extra module entries for JAR files: CLASSPATH."""

        txt = super(JAR, self).make_module_extra()

        for srcfile in self.src:
            srcname = srcfile['name']

            self.log.debug('Checking %s...' % srcname)

            if srcname.endswith('.jar'):
                self.log.debug('Adding %s to classpath' % srcname)
                txt += self.module_generator.prepend_paths('CLASSPATH', [srcname])

        return txt



##
# Copyright 2021-2023 Ghent University
#
# This file is part of EasyBuild,
# originally created by the HPC team of Ghent University (http://ugent.be/hpc/en),
# with support of Ghent University (http://ugent.be/hpc),
# the Flemish Supercomputer Centre (VSC) (https://www.vscentrum.be),
# Flemish Research Foundation (FWO) (http://www.fwo.be/en)
# and the Department of Economy, Science and Innovation (EWI) (http://www.ewi-vlaanderen.be/en).
#
# https://github.com/easybuilders/easybuild
#
# EasyBuild is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation v2.
#
# EasyBuild is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with EasyBuild.  If not, see <http://www.gnu.org/licenses/>.
##
"""
EasyBuild support for building and installing FlexiBLAS, implemented as an easyblock

author: Kenneth Hoste (HPC-UGent)
"""
import os

from easybuild.easyblocks.generic.cmakemake import CMakeMake
from easybuild.framework.easyconfig import CUSTOM
from easybuild.tools import toolchain
from easybuild.tools.build_log import EasyBuildError
from easybuild.tools.config import build_option
from easybuild.tools.environment import setvar
from easybuild.tools.filetools import write_file
from easybuild.tools.run import run_cmd
from easybuild.tools.systemtools import get_shared_lib_ext

IMKL_CONF_TEMPLATE = """[IMKL]
library = libflexiblas_imkl_%(parallel)s_thread.so
[IMKL_SEQ]
library = libflexiblas_imkl_sequential.so
"""


class EB_FlexiBLAS(CMakeMake):
    """Support for building/installing FlexiBLAS."""

    @staticmethod
    def extra_options():
        extra_vars = CMakeMake.extra_options()
        extra_vars.update({
            'blas_auto_detect': [False, "Let FlexiBLAS autodetect the BLAS libraries during configuration", CUSTOM],
            'enable_lapack': [True, "Enable LAPACK support, also includes the wrappers around LAPACK", CUSTOM],
            'flexiblas_default': [None, "Default BLAS lib to set at compile time. If not defined, " +
                                  "the first BLAS lib in backends or the list of dependencies is set as default",
                                  CUSTOM],
            'backends': [None, "List of (build)dependency names to use as BLAS library backends, or " +
                         "'imkl', which does not need to be a (build)dependency." +
                         "If not defined, use the list of dependencies.", CUSTOM],
        })
        extra_vars['separate_build_dir'][0] = True
        return extra_vars

    def __init__(self, *args, **kwargs):
        """Easyblock constructor."""
        super(EB_FlexiBLAS, self).__init__(*args, **kwargs)

        dep_names = [dep['name'] for dep in self.cfg.dependencies()]
        if self.cfg['backends']:
            self.blas_libs = self.cfg['backends'][:]
            # make sure that all listed backends except imkl are (build)dependencies
            if 'imkl' in self.blas_libs and 'imkl' not in dep_names:
                self.blas_libs.remove('imkl')
            backends_nodep = [x for x in self.blas_libs if x not in dep_names]
            if backends_nodep:
                raise EasyBuildError("One or more backends not listed as (build)dependencies: %s",
                                     ', '.join(backends_nodep))
        else:
            build_dep_names = set(dep['name'] for dep in self.cfg.dependencies(build_only=True))
            self.blas_libs = [x for x in dep_names if x not in build_dep_names]

        self.obj_builddir = os.path.join(self.builddir, 'easybuild_obj')

    def configure_step(self):
        """Custom configuration for FlexiBLAS, based on which BLAS libraries are included as dependencies."""

        configopts = {
            'BLAS_AUTO_DETECT': ('OFF', 'ON')[bool(self.cfg['blas_auto_detect'])],
            'LAPACK': ('OFF', 'ON')[bool(self.cfg['enable_lapack'])],
            'FLEXIBLAS_DEFAULT': self.cfg['flexiblas_default'] or self.blas_libs[0],
        }

        supported_blas_libs = ['BLIS', 'NETLIB', 'OpenBLAS', 'imkl']

        # make sure that default backend is a supported library
        flexiblas_default = configopts['FLEXIBLAS_DEFAULT']
        if flexiblas_default not in supported_blas_libs:
            raise EasyBuildError("%s not in list of supported BLAS libs %s", flexiblas_default, supported_blas_libs)

        # make sure that all backends are supported libraries
        unsupported_libs = [x for x in self.blas_libs if x not in supported_blas_libs]
        if unsupported_libs:
            raise EasyBuildError("One or more unsupported libraries used: %s", ', '.join(unsupported_libs))

        # list of BLAS libraries to use is specified via -DEXTRA=...
        configopts['EXTRA'] = ';'.join(self.blas_libs)

        # for each library, we have to specify how to link to it via -DXXX_LIBRARY;
        # see https://github.com/mpimd-csc/flexiblas#setup-with-custom-blas-and-lapack-implementations
        for blas_lib in self.blas_libs:
            key = '%s_LIBRARY' % blas_lib
            if blas_lib == 'imkl':
                # For MKL there is gf_lp64 vs. intel_lp64 and gnu_thread vs. intel_thread (vs. sequential)
                # For gf_lp64 vs intel_lp64 the difference is in the ABI for [sz]dot[uc], which FlexiBLAS
                # can transparenly wrap.
                # gnu_thread vs intel_thread links to libgomp vs. libiomp5 for the OpenMP library.
                mkl_gnu_libs = "mkl_gf_lp64;mkl_gnu_thread;mkl_core;gomp;pthread;m;dl"
                mkl_intel_libs = "mkl_intel_lp64;mkl_intel_thread;mkl_core;iomp5;pthread;m;dl"
                mkl_compiler_mapping = {
                    toolchain.GCC: mkl_gnu_libs,
                    toolchain.INTELCOMP: mkl_intel_libs,
                    toolchain.NVHPC: mkl_intel_libs,
                    toolchain.PGI: mkl_intel_libs,
                }
                comp_family = self.toolchain.comp_family()
                try:
                    configopts[key] = mkl_compiler_mapping[comp_family]
                except KeyError:
                    raise EasyBuildError("Compiler family not supported yet: %s", comp_family)
            else:
                configopts[key] = blas_lib.lower()

        # only add configure options to configopts easyconfig parameter if they're not defined yet,
        # to allow easyconfig to override specifies settings
        for key, value in sorted(configopts.items()):
            opt = '-D%s=' % key
            if opt not in self.cfg['configopts']:
                self.cfg.update('configopts', opt + "'%s'" % value)

        # specify compiler commands with absolute paths, to ensure that RPATH wrapper scripts are used
        if build_option('rpath'):
            self.cfg['abs_path_compilers'] = True

        super(EB_FlexiBLAS, self).configure_step(builddir=self.obj_builddir)

    def install_step(self):
        """Install imkl configuration, found via FLEXIBLAS_LIBRARY_PATH set by imkl module."""

        super(EB_FlexiBLAS, self).install_step()
        if self.cfg['backends'] and 'imkl' in self.cfg['backends'] and 'imkl' not in self.blas_libs:
            if self.toolchain.comp_family() == toolchain.GCC:
                parallel = "gnu"
            else:
                parallel = "intel"
            write_file(os.path.join(self.installdir, 'etc', 'flexiblasrc.d', 'imkl.conf'),
                       IMKL_CONF_TEMPLATE % {'parallel': parallel})

    def test_step(self):
        """Run tests using each of the backends."""

        if build_option('rpath'):
            # inject build location of libflexiblas.so.3 to $LD_LIBRARY_PATH
            # (which is used as a fallback to find libraries not found via the locations listed in the RPATH section),
            # to ensure that test binaries can find the FlexiBLAS library
            ld_library_path = ':'.join([
                os.path.join(self.obj_builddir, 'lib'),
                os.path.join(self.obj_builddir, 'lib64'),
                '$LD_LIBRARY_PATH'
            ])
            self.cfg['pretestopts'] = ('export LD_LIBRARY_PATH="%s" && ' % ld_library_path) + self.cfg['pretestopts']

        # Show verbose output if a test fails
        setvar('CTEST_OUTPUT_ON_FAILURE', 'True')

        # run tests with default backend (NETLIB)
        test_cmd = ' '.join([
            self.cfg['pretestopts'],
            "make test",
            self.cfg['testopts'],
        ])
        run_cmd(test_cmd)

    def sanity_check_step(self):
        """Custom sanity check for FlexiBLAS."""

        shlib_ext = get_shared_lib_ext()

        libs = []

        # libraries in lib/
        top_libs = ['libflexiblas%s.%s' % (x, shlib_ext) for x in ('', '_api', '_mgmt')]
        libs.extend(os.path.join('lib', lf) for lf in top_libs)

        # libraries in lib/flexiblas/
        lower_lib_names = self.blas_libs + ['hook_dummy', 'hook_profile']
        if self.cfg['enable_lapack']:
            lower_lib_names += ['fallback_lapack']
        lower_libs = ['libflexiblas_%s.%s' % (x.lower(), shlib_ext) for x in lower_lib_names]
        libs.extend(os.path.join('lib', 'flexiblas', lf) for lf in lower_libs)

        # include files in include/flexiblas
        flexiblas_includes = ['flexiblas_api.h', 'flexiblas_mgmt.h', 'cblas.h']
        if self.cfg['enable_lapack']:
            flexiblas_includes += ['lapack.h']
        includes = [os.path.join('include', 'flexiblas', x) for x in flexiblas_includes]

        custom_paths = {
            'files': [os.path.join('bin', 'flexiblas'), os.path.join('etc', 'flexiblasrc')] + includes + libs,
            'dirs': [os.path.join('etc', 'flexiblasrc.d'), os.path.join('share', 'man')],
        }

        custom_commands = [
            "flexiblas --help",
            "flexiblas list",
        ]

        # make sure that each BLAS library is supported by FlexiBLAS by checking output of 'flexiblas list'
        for blas_lib in self.blas_libs:
            custom_commands.append("flexiblas list | grep %s" % blas_lib.upper())

        super(EB_FlexiBLAS, self).sanity_check_step(custom_paths=custom_paths, custom_commands=custom_commands)

    def make_module_req_guess(self):
        """Customize CPATH for FlexiBLAS."""

        guesses = super(EB_FlexiBLAS, self).make_module_req_guess()

        guesses.update({'CPATH': [os.path.join('include', 'flexiblas')]})

        return guesses
